from rolldecayestimators import ikeda_speed as ikeda
from numpy import sqrt, pi
import numpy as np
import pandas as pd
from numpy.testing import assert_allclose
import matplotlib.pyplot as plt
import pytest

ScaleF =  1 #%/29.565;                # Scale Factor [-]
visc =   1.15*10**-6;                 # [m2/s], kinematic viscosity
Cb   =   0.61;                        # Block coeff
L    =   220*ScaleF;                  # Length
vcg  =   14.4*ScaleF;                 # roll axis (vertical centre of gravity) [m]
vcg  =   14.9*ScaleF;                 # roll axis (vertical centre of gravity) [m]
B    =   32.26*ScaleF;                # Breadth of hull [m]
d    =   9.5*ScaleF;                  # Draught of hull [m]
A    =   0.93*B*d;                    # Area of cross section of hull [m2]
bBK  =   0.4*ScaleF;                  # breadth of Bilge keel [m] !!(=height???)
R    =   5*ScaleF;                    # Bilge Radis
g    =   9.81;
C_mid=   0.93;

OG = -1*(vcg-d)#*0.8;                    # distance from roll axis to still water level
Ho = B/(2*d);                  # half breadth to draft ratio
ra   = 1025;                   # density of water

#locals
LBK  = L/4;                    # Approx
disp = L*B*d*Cb;               # Displacement

# variables!!
T=27.6*sqrt(ScaleF); wE   = 2*pi*1/T;        # circular frequency
fi_a =   10*pi/180;            # roll amplitude !!rad??
V    =   0;                  # Speed

ND_factorB = sqrt(B/(2*g))/(ra*disp*(B**2));   # Nondimensiolizing factor of B44
ND_factor = wE/(2*ra*disp*g*(16*ScaleF-vcg));   # Nondimensiolizing factor of B44

def test_Bw0():
    Bw0=ikeda.Bw0_faust(w=0.227651641564478)
    assert_allclose(Bw0,1.763939936010376e+06, rtol=0.001)

def test_bw44_V0():
    V = 0  ## Ship speed
    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency


    bw44 = ikeda.Bw_faust(w=wE, V=V, d=d)

    assert_allclose(bw44, 1.523005619852937e+07, rtol=0.001)

def test_bilge_keel():
    V = 0  ## Ship speed
    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency
    fi_a = 10 * pi / 180;  # roll amplitude !!rad??
    Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0 = ikeda.bilge_keel(w=wE, fi_a=fi_a, V=V, B=B, d=d, A=A, bBK=bBK, R=R, g=g, OG=OG, Ho=Ho, ra=ra)

    assert_allclose(Bp44BK_N0, 3.509033728796537e+05, rtol=0.001)
    assert_allclose(Bp44BK_H0, 1.057765842625329e+06, rtol=0.001)
    assert_allclose(B44BK_L, 2.607161328890767e+05, rtol=0.001)
    assert_allclose(B44BKW0, 0.011960200569156, rtol=0.001)

@pytest.mark.skip('This one does not work yet')
def test_friction():
    V = 0  ## Ship speed
    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency
    fi_a = 10 * pi / 180;  # roll amplitude !!rad??
    B44F = ikeda.frictional(w=wE, fi_a=fi_a, V=V, B=B, d=d, OG=OG, ra=ra, Cb=Cb, L=L, visc=visc)
    assert_allclose(B44F, 5.794039754129194e+06, rtol=0.001)

def test_hull_lift():
    V = 10  ## Ship speed
    B44L = ikeda.hull_lift(V=V, B=B, d=d, OG=OG, ra=ra, L=L, A=A)
    assert_allclose(B44L, 1.734463413980598e+08, rtol=0.001)

@pytest.mark.skip('This one does not work yet')
def test_calculate_B44():
    V = 10  ## Ship speed
    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency
    fi_a = 10 * pi / 180;  # roll amplitude !!rad??
    Bw0 = ikeda.Bw0_faust(w=wE)
    B44,BW44,B44_BK,B44F,B44L = ikeda.calculate_B44(w=wE, V=V, d=d, Bw0=Bw0, fi_a=fi_a,  B=B,  A=A, bBK=bBK, R=R, OG=OG,
                                                    Ho=Ho, ra=ra, Cb=Cb, L=L, LBK=LBK, visc = visc, g=g, Bw_div_Bw0_max=np.inf)

    assert_allclose(BW44, 44060267.568187, rtol=0.001)
    assert_allclose(B44_BK, 79564994.97049402, rtol=0.001)
    assert_allclose(B44F, 10537244.373232083, rtol=0.001)
    assert_allclose(B44L, 173446341.3980598, rtol=0.001)

@pytest.mark.skip('This one does not work yet')
def test_calculate_B44_speeds():

    V = np.arange(0,10+1)

    T = 27.6 * sqrt(ScaleF);
    wE = 2 * pi * 1 / T;  # circular frequency
    fi_a = 10 * pi / 180;  # roll amplitude !!rad??
    Bw0 = ikeda.Bw0_faust(w=wE)

    inputs = pd.DataFrame(index=V)
    inputs['V']=V
    inputs['w']=wE
    inputs['fi_a']=fi_a
    inputs['Bw0']=Bw0
    inputs['d']=d
    inputs['B']=B
    inputs['A']=A
    inputs['BKB']=bBK
    inputs['R_b']=R
    inputs['OG']=OG
    inputs['Ho']=Ho
    inputs['rho']=ra
    inputs['Cb']=Cb
    inputs['L']=L
    inputs['LBK']=LBK
    inputs['visc']=visc
    inputs['g']=g

    results = inputs.apply(func=ikeda.calculate_B44_series, Bw_div_Bw0_max=np.inf, axis=1)

    # This one is no longer valid
    # Wave:
    #BW44_matlab = [15230056.1985294,15245650.7598695,15312025.5743705,15551094.8502096,16275924.5707451,18116404.2772792,22004594.1044359,28771214.8371744,38312799.5815772,48887362.9192075,57522599.4248384]
    #assert_allclose(results['B_W'], BW44_matlab, rtol=0.001)


    # B44_BK
    B44_BK_matlab = [77737522.9856631,77764104.9949679,77935523.2118694,78129653.2817479,78330461.4926025,78534065.8999322,78739095.8600784,78944948.7197441,79151318.9709843,79358035.4695565,79564994.9704940]
    assert_allclose(results['B_BK'], B44_BK_matlab, rtol=0.001)

    # B44F:
    B44F_matlab = [5794039.75412919,6268360.21603948,6742680.67794977,7217001.13986006,7691321.60177035,8165642.06368064,8639962.52559093,9114282.98750122,9588603.44941150,10062923.9113218,10537244.3732321]
    assert_allclose(results['B_F'], B44F_matlab, rtol=0.001)

    # Lift:
    B44L_matlab = [0,17344634.1398060,34689268.2796119,52033902.4194179,69378536.5592239,86723170.6990299,104067804.838836,121412438.978642,138757073.118448,156101707.258254,173446341.398060]
    assert_allclose(results['B_L'], B44L_matlab, rtol=0.001)

    fig,ax=plt.subplots()
    result_nondim = results*ND_factor
    result_nondim.plot.area(y = ['B_BK','B_L','B_W','B_F'], ax=ax)
    plt.show()




import numpy as np
import pandas as pd
from scipy.integrate import odeint
from rolldecayestimators import DirectEstimator
from rolldecayestimators.symbols import *
from rolldecayestimators.substitute_dynamic_symbols import lambdify

dGM = sp.symbols('dGM')
lhs = phi_dot_dot + 2*zeta*omega0*phi_dot + omega0**2*phi+(dGM*phi*sp.Abs(phi)) + d*sp.Abs(phi_dot)*phi_dot
roll_diff_equation = sp.Eq(lhs=lhs, rhs=0)
acceleration = sp.Eq(lhs=phi, rhs=sp.solve(roll_diff_equation, phi.diff().diff())[0])
calculate_acceleration = lambdify(acceleration.rhs)

# Defining the diff equation for this estimator:

class DirectEstimatorImproved(DirectEstimator):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """

    @staticmethod
    def estimator(df, omega0, zeta, dGM, d):
        phi = df['phi']
        phi1d = df['phi1d']

        phi2d = calculate_acceleration(omega0=omega0, phi=phi, phi1d=phi1d, zeta=zeta, dGM=dGM, d=d)

        return phi2d

    @staticmethod
    def roll_decay_time_step(states, t, omega0, zeta, dGM, d):
        # states:
        # [phi,phi1d]

        phi_old = states[0]
        p_old = states[1]

        phi1d = p_old
        phi2d = calculate_acceleration(omega0=omega0, phi=phi_old, phi1d=p_old, zeta=zeta, dGM=dGM, d=d)

        d_states_dt = np.array([phi1d, phi2d])

        return d_states_dt

    def simulate(self, t: np.ndarray, phi0: float, phi1d0: float, omega0: float, zeta: float, dGM: float,
                 d: float) -> pd.DataFrame:
        """
        Simulate a roll decay test using the quadratic method.
        :param t: time vector to be simulated [s]
        :param phi0: initial roll angle [rad]
        :param phi1d0: initial roll speed [rad/s]
        :param omega0: roll natural frequency[rad/s]
        :param d: quadratic roll damping [-]
        :param zeta:linear roll damping [-]
        :return: pandas data frame with time series of 'phi' and 'phi1d'
        """

        states0 = [phi0, phi1d0]
        args = (
            omega0,
            zeta,
            dGM,
            d,
        )
        states = odeint(func=self.roll_decay_time_step, y0=states0, t=t, args=args)

        df = pd.DataFrame(index=t)

        df['phi'] = states[:, 0]
        df['phi1d'] = states[:, 1]

        return df



"""
This is a module to be used as a reference for building other modules
"""
import numpy as np
import pandas as pd
from scipy.integrate import odeint
from rolldecayestimators.substitute_dynamic_symbols import lambdify
from rolldecayestimators.symbols import *
import inspect
from scipy.optimize import curve_fit

from rolldecayestimators.direct_estimator import DirectEstimator


lhs = phi_dot_dot + 2*zeta*omega0*phi_dot + omega0**2*phi
roll_diff_equation = sp.Eq(lhs=lhs,rhs=0)
acceleration = sp.Eq(lhs=phi, rhs=sp.solve(roll_diff_equation, phi.diff().diff())[0])
calculate_acceleration = lambdify(acceleration.rhs)


class DirectLinearEstimator(DirectEstimator):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    demo_param : str, default='demo_param'
        A parameter used for demonstation of how to pass and store paramters.
    """
    # Defining the diff equation for this estimator:
    rhs = -phi_dot_dot/(omega0**2) - 2*zeta/omega0*phi_dot
    roll_diff_equation = sp.Eq(lhs=phi, rhs=rhs)
    acceleration = sp.Eq(lhs=phi, rhs=sp.solve(roll_diff_equation, phi.diff().diff())[0])
    functions = {'acceleration':lambdify(acceleration.rhs)}

    @classmethod
    def load(cls, omega0:float, zeta:float, X=None):
        """
        Load data and parameters from an existing fitted estimator

        A_44 is total roll intertia [kg*m**2] (including added mass)

        Parameters
        ----------
        omega0:
            roll natural frequency[rad/s]
        zeta:
            linear roll damping [-]

        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        data = {
            'omega0': omega0,
            'zeta': zeta,
        }

        return super(cls, cls)._load(data=data, X=X)



import numpy as np
import sympy as sp
import pandas as pd

from rolldecayestimators import equations
from rolldecayestimators import symbols
from rolldecayestimators import lambdas

from rolldecayestimators.substitute_dynamic_symbols import lambdify

lambda_B_1_zeta = lambdify(sp.solve(equations.B_1_zeta_eq, symbols.B_1)[0])

eq_phi1d = sp.Eq(symbols.phi_dot_dot,
                 sp.solve(equations.roll_decay_equation_cubic_A, symbols.phi_dot_dot)[0])

accelaration_lambda = lambdify(sp.solve(eq_phi1d, symbols.phi_dot_dot)[0])


def find_peaks(df_state_space):
    df_state_space['phi_deg'] = np.rad2deg(df_state_space['phi'])

    mask = (np.sign(df_state_space['phi1d']) != np.sign(np.roll(df_state_space['phi1d'], -1))
            )
    mask[0] = False
    mask[-1] = False

    df_max = df_state_space.loc[mask].copy()
    df_max['id'] = np.arange(len(df_max)) + 1
    return df_max


def calculate_decrements(df_amplitudes):
    ## Calculate decrements:
    df_decrements = pd.DataFrame()
    for i in range(len(df_amplitudes) - 2):
        s1 = df_amplitudes.iloc[i]
        s2 = df_amplitudes.iloc[i + 2]
        decrement = s1 / s2
        decrement.name = s1.name
        df_decrements = df_decrements.append(decrement)

    return df_decrements


def calculate_zeta(df_decrements):
    zeta_n = 1 / (2 * np.pi) * np.log(df_decrements['phi'])
    return zeta_n

def calculate_B(zeta_n, A_44, omega0):

    ## Convert to B damping [Nm*s]
    B = lambda_B_1_zeta(A_44=A_44, omega0=omega0, zeta=zeta_n)
    B = np.array(B, dtype=float)
    return B

def estimate_amplitude(phi):
    phi = np.abs(phi)
    phi_a = (phi + np.roll(phi, -1) + np.roll(phi, -2)) / 3
    return phi_a


import pytest
import numpy as np
from numpy.testing import assert_almost_equal, assert_array_almost_equal_nulp

from rolldecayestimators.simplified_ikeda_class import SimplifiedIkeda
from rolldecayestimators.simplified_ikeda import calculate_roll_damping

L_div_B = 6.0
BD = 4.0
CB = 0.65
CMID = 0.98
PHI = 10
BBKB = 0.025
LBKL = 0.2
OGD = -0.2
OMEGAHAT = 0.6
LPP = 300

Beam = LPP / L_div_B
DRAFT = Beam / BD

lBK = LPP * LBKL
bBK = Beam * BBKB
OMEGA = OMEGAHAT / (np.sqrt(Beam / 2 / 9.81))

OG = DRAFT * OGD
kg = DRAFT - OG
volume = LPP * DRAFT * Beam * CB

B44HAT1, BFHAT1, BWHAT1, BEHAT1, BBKHAT1, BLHAT1 = calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, lBK, bBK, OMEGA,
                                                                          DRAFT)
decimal=5

def test_simplified_ikeda():


    si = SimplifiedIkeda(V=0, w=OMEGA, fi_a=np.deg2rad(PHI), beam=Beam, lpp=LPP, kg=kg, volume=volume, draught=DRAFT, A0=CMID,
                         BKL=lBK, BKB=bBK, visc=1.14e-6)

    B44HAT = si.calculate_B44()
    BFHAT = si.calculate_B_F()
    BWHAT = si.calculate_B_W()
    BBKHAT = si.calculate_B_BK()
    BLHAT = si.calculate_B_L()


    assert_almost_equal(BFHAT, BFHAT1, decimal=decimal)
    assert_almost_equal(BWHAT, BWHAT1, decimal=decimal)
    assert_almost_equal(BBKHAT, BBKHAT1, decimal=decimal)
    assert_almost_equal(BLHAT, BLHAT1, decimal=decimal)
    assert_almost_equal(B44HAT, B44HAT1, decimal=decimal)


def test_simplified_ikeda_scale():

    V = 10

    si = SimplifiedIkeda(V=V, w=OMEGA, fi_a=np.deg2rad(PHI), beam=Beam, lpp=LPP, kg=kg, volume=volume, draught=DRAFT, A0=CMID,
                         BKL=lBK, BKB=bBK, visc=1.14e-6)

    scale_factor = 100
    V_m = V/np.sqrt(scale_factor)
    w_m= OMEGA*np.sqrt(scale_factor)
    beam_m = Beam/scale_factor
    lpp_m = LPP/scale_factor
    kg_m = kg/scale_factor
    draught_m = DRAFT/scale_factor
    lBK_m = lBK/scale_factor
    bBK_m = bBK/scale_factor
    volume_m = volume/(scale_factor**3)

    si_model = SimplifiedIkeda(V=V_m, w=w_m, fi_a=np.deg2rad(PHI), beam=beam_m, lpp=lpp_m, kg=kg_m, volume=volume_m, draught=draught_m,
                               A0=CMID,
                               BKL=lBK_m, BKB=bBK_m, visc=1.14e-6)

    assert_almost_equal(si.calculate_B_W(), si_model.calculate_B_W(), decimal=decimal)
    assert_almost_equal(si.calculate_B_BK(), si_model.calculate_B_BK(), decimal=decimal)
    assert_almost_equal(si.calculate_B_E(), si_model.calculate_B_E(), decimal=decimal)
    assert_almost_equal(si.calculate_B_L(), si_model.calculate_B_L(), decimal=decimal)

    # The friction is not froude scaled:
    with pytest.raises(AssertionError):
        assert_almost_equal(si.calculate_B_F(), si_model.calculate_B_F(), decimal=decimal)


def test_simplified_ikeda_vector():

    N=10
    w = np.ones(N)*OMEGA
    si = SimplifiedIkeda(V=0, w=w, fi_a=np.deg2rad(PHI), beam=Beam, lpp=LPP, kg=kg, volume=volume, draught=DRAFT, A0=CMID,
                         BKL=lBK, BKB=bBK, visc=1.14e-6)

    B44HAT = si.calculate_B44()

    assert_almost_equal(B44HAT, np.ones(N)*B44HAT1, decimal=decimal)

def test_simplified_ikeda_vector2():

    N=10
    fi_a = np.deg2rad(np.linspace(1, 15, N))

    si = SimplifiedIkeda(V=0, w=OMEGA, fi_a=fi_a, beam=Beam, lpp=LPP, kg=kg, volume=volume, draught=DRAFT, A0=CMID,
                         BKL=lBK, BKB=bBK, visc=1.14e-6)

    B44HAT = si.calculate_B44()

    a = 1




"""
Equations to predict eddy damping according to:

[1]
Ikeda, Y.,
1978. On eddy making component of roll damping force on naked hull. University of Osaka Prefacture,
Department of Naval Architecture, Japan, Report No. 00403,
Published in: Journal of Society of Naval Architects of Japan, Volume 142.

[2]
ITTC, 2011. ITTC – Recommended Procedures Numerical Estimation of Roll Damping.


"""

import sympy as sp
from rolldecayestimators.symbols import *
from sympy import pi,sqrt

eq_B_E_star_hat = sp.Eq(B_E0_hat,
                        8/(3*pi)*B_E_star_hat  # (6) (part 1)
                        )

eq_B_E0_hat = sp.Eq(B_E0_hat,
                    4*L_pp*T_s**4*omega_hat*phi_a/(3*pi*Disp*beam**2)*C_r  # (6) (part 2)
                    )
eq_volume = sp.Eq(Disp,
                  T_s*B_s*sigma*L_pp,
                  )
solution = sp.solve([eq_B_E0_hat, eq_volume, eq_B_E_star_hat],
                    C_r,Disp,B_E_star_hat,
                    dict=True
                    )
eq_C_r_2 = sp.Eq(C_r, solution[0][C_r])

eq_C_r = sp.Eq(C_r,
               2/(rho*T_s**2)*((1 - f_1*R_b/T_s)*(1 - OG/T_s - f_1*R_b/T_s) + f_2*(H_0 - f_1*R_b/T_s)**2)*P_m/3  # (10)
               )

eq_P_m = sp.Eq(P_m,
               3*1/2*rho*r_max**2*C_P*sp.Abs(phi.diff())*phi.diff().diff()  # (13)
              )

eq_B_E0s = sp.Eq(B_E0s,
                 4 * rho * T_s**4*omega*phi_a * C_r / (3 * pi) # (2.19) [2]
            )

eq_B_E0 = sp.Eq(B_E0,
                sp.Integral(B_E0s,(x_s,AP,FP))
     )

eq_R_b = sp.Eq(R_b,
               sqrt((B_s*T_s*(1-sigma))/(1-pi/4))  # Martin R_b approximations
               )


eq_B_star_hat = sp.Eq(B_star_hat,
                      B_E_star_hat +  B_W_star_hat + B_F_star_hat)

## Lewis:
eq_D_1 = sp.Eq(D_1,
                (3 + 4 * sigma / pi) + (1 - 4 * sigma / pi) * ((H_0 - 1) / (H_0 + 1)) ** 2
                )

eq_a_3 = sp.Eq(a_3,
               (-D_1 + 3 + sqrt(9 - 2 * D_1)) / D_1
               )
eq_a_1 = sp.Eq(a_1,
               (1 + a_3) * (H_0 - 1) / (H_0 + 1)
               )


import numpy as np
import pandas as pd

from rolldecayestimators.ikeda import Ikeda
import rolldecayestimators.ikeda
import rolldecayestimators.simplified_ikeda as si
from rolldecayestimators import ikeda_speed

class SimplifiedIkeda(Ikeda):
    """
        Class that helps with running various calculations known as the "Simplified Ikedas method" to predict ship roll damping.
        The idea with this class is that it can be subclassed so that varius roll damping contributions can be replaced or
        removed in order to study different "flavors" of the Ikeda methods (I have seen a lot of different implementations).
        """

    def __init__(self, V: np.ndarray, w: np.ndarray, fi_a: float, beam: float, lpp: float,
                 kg: float, volume: float, draught: float, A0: float, BKL:float, BKB:float,
                 g=9.81, rho=1000.0, visc=1.15 * 10 ** -6, **kwargs):
        """
                Manually specify the inputs to the calculations.
                Note: Some inputs need to be specified here, but others can be defined in other ways
                (by importing a ship geometry etc.)

                Parameters
                ----------
                V
                    ship speed [m/s]
                w
                    roll frequency [rad/s]
                fi_a
                    roll amplitude [rad]
                beam
                    ship beam [m]
                lpp
                    ship perpedicular length [m]
                kg
                    vertical centre of gravity [m] (positive upward)
                volume
                    ship displaced volume [m3]
                draught
                    ship draught [m]
                A0
                    mid section coefficient (A0 = A_mid/(B*T))
                BKL
                    length bilge keel [m] (=0 --> no bilge keel)
                BKB
                    height bilge keel [m] (=0 --> no bilge keel)
                g
                    gravity [m/s2]
                rho
                    water density [kg/m3]


                Returns
                -------
                None
                """
        self.V = V
        self.g = g
        self.w = w
        self.fi_a = fi_a
        self.beam = beam
        self.lpp = lpp
        self.kg = kg
        self.volume = volume
        self.lBK = BKL
        self.bBK = BKB
        self.rho = rho
        self.visc = visc

        self._A0=A0
        self._draught=draught



        #B_W0: pd.Series

    @property
    def A0(self):
        return self._A0

    @property
    def draught(self):
        return self._draught

    def calculate_B44(self):
        """
        Calculate total roll damping

        Returns
        -------
        B_44_hat : ndarray
            Nondimensioal total roll damping [-]
        """

        B_44 = (self.calculate_B_W() +
                self.calculate_B_F() +
                self.calculate_B_E() +
                self.calculate_B_L() +
                self.calculate_B_BK()
                )
        return B_44

    def calculate_B_W0(self):
        """
        Calculate roll wave damping at zero speed

        Returns
        -------
        B_W0_hat : ndarray
            Nondimensional roll wave damping at zero speed [-]

        """
        B_W0_hat =  si.calculate_B_W0(BD=self.BD, CB=self.Cb, CMID=self.A0, OGD=self.OGD, OMEGAHAT=self.w_hat)
        return B_W0_hat

    def calculate_B_W(self, Bw_div_Bw0_max=np.inf):
        """
        Calculate roll wave damping at speed

        Returns
        -------
        B_W_hat : ndarray
            Nondimensional roll wave damping at speed [-]

        """
        B_W0_hat = self.calculate_B_W0()
        Bw_div_Bw0 = self.calculate_Bw_div_Bw0()
        B_W_hat = B_W0_hat*Bw_div_Bw0
        return B_W_hat

    def calculate_B_F(self):
        """
        Calculate skin friction damping

        Returns
        -------
        B_F_hat : ndarray
            Nondimensional skin friction damping [-]

        """
        B_F_hat = si.calculate_B_F(BD=self.BD, BRTH=self.beam, CB=self.Cb, DRAFT=self.draught, KVC=self.visc,
                                   LPP=self.lpp, OGD=self.OGD, OMEGA=self.w, PHI=np.rad2deg(self.fi_a))
        return B_F_hat

    def calculate_B_E(self):
        """
        Calculate bilge eddy damping

        Returns
        -------
        B_E_hat : ndarray
            Nondimensional eddy damping [-]

        """
        B_E_hat = si.calculate_B_E(BD=self.BD, CB=self.Cb, CMID=self.A0, OGD=self.OGD, OMEGAHAT=self.w_hat,
                                   PHI=np.rad2deg(self.fi_a))
        return B_E_hat

    def calculate_B_BK(self):
        """
        Calculate bilge keel damping

        Returns
        -------
        B_BK_hat : ndarray
            Nondimensional bilge keel damping [-]
        """
        B_BK_hat = si.calculate_B_BK(BBKB=self.bBK/self.beam, BD=self.BD, CB=self.Cb, CMID=self.A0, LBKL=self.lBK/self.lpp, OGD=self.OGD,
                                     OMEGAHAT=self.w_hat, PHI=np.rad2deg(self.fi_a))
        return B_BK_hat


class SimplifiedIkedaBK2(SimplifiedIkeda):

    def calculate_B_BK(self):
        """
        Calculate bilge keel damping

        Returns
        -------
        B_BK_hat : ndarray
            Bilge keel damping [-]

        """

        if np.any(~(self.bBK == 0) & (self.lBK == 0)):
            raise rolldecayestimators.ikeda.BilgeKeelError('BKB is 0 but BKL is not!')
            return 0.0

        if isinstance(self.R, np.ndarray):
            index = int(len(self.R) / 2)  # Somewhere in the middle of the ship
            R = self.R[index]
        else:
            R = self.R

        Bp44BK_N0, Bp44BK_H0, B44BK_L, B44BKW0 = ikeda_speed.bilge_keel(w=self.w, fi_a=self.fi_a, V=self.V, B=self.beam,
                                                                        d=self.draught, A=self.A_mid,
                                                                        bBK=self.bBK, R=R, g=self.g, OG=self.OG,
                                                                        Ho=self.Ho, ra=self.rho)

        B44BK_N0 = Bp44BK_N0 * self.lBK
        B44BK_H0 = Bp44BK_H0 * self.lBK
        B44_BK = B44BK_N0 + B44BK_H0 + B44BK_L

        B44_BK=rolldecayestimators.ikeda.array(self.B_hat(B44_BK))
        mask = ((self.lBK == 0) | (pd.isnull(self.lBK)))
        B44_BK[mask] = 0

        return B44_BK


class SimplifiedIkedaABS(SimplifiedIkeda):

    def calculate_B_W0(self):
        """
        Calculate roll wave damping at zero speed

        Returns
        -------
        B_W0_hat : ndarray
            Nondimensional roll wave damping at zero speed [-]

        """
        return np.abs(super().calculate_B_W0())

    def calculate_B_W(self, Bw_div_Bw0_max=np.inf):
        """
        Calculate roll wave damping at speed

        Returns
        -------
        B_W_hat : ndarray
            Nondimensional roll wave damping at speed [-]

        """
        return np.abs(super().calculate_B_W())

    def calculate_B_F(self):
        """
        Calculate skin friction damping

        Returns
        -------
        B_F_hat : ndarray
            Nondimensional skin friction damping [-]

        """
        return np.abs(super().calculate_B_F())

    def calculate_B_E(self):
        """
        Calculate bilge eddy damping

        Returns
        -------
        B_E_hat : ndarray
            Nondimensional eddy damping [-]

        """
        return np.abs(super().calculate_B_E())

    def calculate_B_BK(self):
        """
        Calculate bilge keel damping

        Returns
        -------
        B_BK_hat : ndarray
            Nondimensional bilge keel damping [-]
        """
        return np.abs(super().calculate_B_BK())


from sklearn.metrics import r2_score
import pandas as pd
pd.options.display.max_rows = 999
pd.options.display.max_columns = 999
pd.set_option("display.max_columns", None)
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
import os.path

from rolldecayestimators.substitute_dynamic_symbols import lambdify,run


from sklearn.linear_model import LinearRegression
from sympy.parsing.sympy_parser import parse_expr
import sympy as sp
from rolldecayestimators import symbols
import dill
dill.settings['recurse'] = True

class Polynom(BaseEstimator):
    """ Estimator that wrappes a model pipline and convert it to a SymPy polynomial expression

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    model  : sklearn.pipeline.Pipeline,
        The model should contain:
        model['polynomial_feature'] : sklearn.feature_selection.SelectKBest
        model['variance_treshold'] : sklearn.feature_selection.VarianceThreshold
        model['select_k_best'] : sklearn.feature_selection.SelectKBest

    columns : list
        a list with the column names from the original dataframe.

    y_symbol : sympy.Symbol
        y_symbol = poly(....)

    """


    def __init__(self, model:Pipeline, columns:list, y_symbol:sp.Symbol):

        self.polynomial_features = model['polynomial_feature']
        self.variance_treshold = model['variance_treshold']
        self.select_k_best = model['select_k_best']

        self.feature_names = np.array(self.polynomial_features.get_feature_names())
        self.polynomial_regression = LinearRegression()
        self.columns = columns
        self.feature_eqs = self.define_feature_equations()
        self.y_symbol = y_symbol

    def fit(self, X, y):
        self.X = X
        result = self.polynomial_regression.fit(X=self.good_X(X), y=y)
        self.equation = self.get_equation()
        self.lamda = lambdify(self.equation.rhs)
        return result

    def score(self, X, y):
        y_pred = self.predict(X=X)
        return r2_score(y_true=y, y_pred=y_pred)

    def predict(self, X):

        if isinstance(X,dict):
            return run(self.lamda, X)

        if isinstance(X,pd.Series):
            return run(self.lamda, X)

        if not isinstance(X, pd.DataFrame):
            assert X.shape[1] == len(self.columns)
            X = pd.DataFrame(data=X, columns=self.columns)

        return run(self.lamda, X)

    @property
    def good_index(self):
        feature_names_index = pd.Series(self.feature_names)
        mask_treshold = self.variance_treshold.get_support()
        mask_select_k_best = self.select_k_best.get_support()
        return feature_names_index[mask_treshold][mask_select_k_best]

    def good_X(self, X):
        X2 = self.polynomial_features.transform(X)
        return X2[:, self.good_index.index]  # Only the good stuff

    def define_sympy_symbols(self):
        self.sympy_symbols = {key: getattr(symbols, key) for key in self.columns}

    def define_parameters(self):
        self.parameters = [self.sympy_symbols[key] for key in self.columns]

    def define_feature_equations(self):

        self.define_sympy_symbols()
        self.define_parameters()

        xs = {'x%i' % i: name for i, name in enumerate(self.columns)}
        xs_sympy = {sp.Symbol(key): self.sympy_symbols[value] for key, value in xs.items()}

        feature_eqs = [1.0, ]

        for feature in self.feature_names[1:]:
            s_eq = feature

            s_eq = s_eq.replace(' ', '*')
            s_eq = s_eq.replace('^', '**')

            sympy_eq = parse_expr(s_eq)
            sympy_eq = sympy_eq.subs(xs_sympy)
            feature_eqs.append(sympy_eq)

        return np.array(feature_eqs)

    @property
    def good_feature_equations(self):
        return self.feature_eqs[self.good_index.index]

    def get_equation(self):
        rhs = 0
        for good_feature_equation, coeff in zip(self.good_feature_equations,
                                                self.polynomial_regression.coef_):
            rhs += coeff * good_feature_equation
        rhs += self.polynomial_regression.intercept_

        return sp.Eq(self.y_symbol, rhs)

    def save(self, file_path:str):

        if not os.path.splitext(file_path)[-1]:
            file_path+='.sym'

        dill.dump(self, open(file_path, 'wb'))

    @classmethod
    def load(cls, file_path:str):
        with open(file_path, 'rb') as file:
            polynom = dill.load(file)

        polynom.equation = polynom.get_equation()
        polynom.lamda = lambdify(polynom.equation.rhs)

        return polynom


import pytest
import numpy as np

from sklearn.utils.testing import assert_array_equal
from sklearn.utils.testing import assert_allclose

from rolldecayestimators import DirectEstimator, CutTransformer
from rolldecayestimators.transformers import OffsetTransformer
from rolldecayestimators.tests.test_cubic import simulate
import matplotlib.pyplot as plt

@pytest.fixture
def data():
    parameters = {
        'B_1A': 0.3,
        'B_2A': 0.0,
        'B_3A': 0.0,
        'C_1A': 10.0,
        'C_3A': 0.0,
        'C_5A': 0.0,
    }

    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 15, 0.1)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    return X

def test_offset_transformer(data):
    X = data

    X_offset = X.copy()
    phi_offset = np.deg2rad(1.0)
    t = X_offset.index
    delta_t = t[-1] - t[0]
    d_phi = phi_offset/delta_t
    X_offset['phi']=X_offset['phi'] + d_phi*t

    trans = OffsetTransformer()

    trans.fit(X=X)

    X_trans = trans.transform(X=X)

    fig,ax=plt.subplots()
    X.plot(y='phi', ax=ax, label='true')
    X_offset.plot(y='phi', ax=ax, label='model test')
    X_trans.plot(y='phi', ax=ax, style='--', label='offset removed')
    ax.legend()
    plt.show()

    assert_allclose(X_trans['phi'], X['phi'], atol=0.01)


import numpy as np
import pandas as pd
from rolldecayestimators import lambdas

def sample_increase(X, increase=5):
    N = len(X) * increase
    t_interpolated = np.linspace(X.index[0], X.index[-1], N)
    X_interpolated = pd.DataFrame(index=t_interpolated)

    for key, values in X.items():
        X_interpolated[key] = np.interp(t_interpolated, values.index, values)

    return X_interpolated

def get_peaks(X:pd.DataFrame, key='phi1d')->pd.DataFrame:
    """
    Find the peaks in the signal by finding zero roll angle velocity

    Parameters
    ----------
    X
        DataFrame with roll signal as "phi"
    key = 'phi1d'

    Returns
    -------
    Dataframe with rows from X where phi1d is close to 0.

    """

    phi1d = np.array(X[key])

    index = np.arange(0, len(X.index))
    index_later = np.roll(index, shift=-1)
    index_later[-1] = index[-1]
    mask = (
            ((phi1d[index] > 0) &
             (phi1d[index_later] < 0)) |
            ((phi1d[index] < 0) &
             (phi1d[index_later] > 0))
    )

    index_first = index[mask]
    index_second = index[mask] + 1

    # y = m + k*x
    # k = (y2-y1)/(x2-x1)
    # m = y1 - k*x1
    # y = 0 --> x = -m/k
    X_1 = X.iloc[index_first].copy()
    X_2 = X.iloc[index_second].copy()
    rows, cols = X_1.shape

    x1 = np.array(X_1.index)
    x2 = np.array(X_2.index)
    y1 = np.array(X_1['phi1d'])
    y2 = np.array(X_2['phi1d'])
    k = (y2 - y1) / (x2 - x1)
    m = y1 - k * x1
    x = -m / k

    X_1 = np.array(X_1)
    X_2 = np.array(X_2)

    factor = (x - x1) / (x2 - x1)
    factor = np.tile(factor, [cols, 1]).T
    X_zero = X_1 + (X_2 - X_1) * factor

    X_zerocrossings = pd.DataFrame(data=X_zero, columns=X.columns, index=x)

    return X_zerocrossings

def calculate_amplitudes(X_zerocrossings):

    X_amplitudes = pd.DataFrame()
    for i in range(len(X_zerocrossings) - 1):
        s1 = X_zerocrossings.iloc[i]
        s2 = X_zerocrossings.iloc[i + 1]

        amplitude = (s2 - s1).abs()
        amplitude.name = (s1.name + s2.name)/2  # mean time
        X_amplitudes = X_amplitudes.append(amplitude)

    X_amplitudes['phi']/=2
    X_amplitudes['phi_a'] = X_amplitudes['phi']

    return X_amplitudes

def calculate_amplitudes_and_damping(X:pd.DataFrame):
    X_interpolated = sample_increase(X=X)
    X_zerocrossings = get_peaks(X=X_interpolated)
    X_amplitudes = calculate_amplitudes(X_zerocrossings=X_zerocrossings)
    X_amplitudes = calculate_damping(X_amplitudes=X_amplitudes)
    T0 = 2*X_amplitudes.index
    X_amplitudes['omega0'] = 2 * np.pi/np.gradient(T0)
    #X_amplitudes['time'] = np.cumsum(X_amplitudes.index)
    return X_amplitudes

def calculate_damping(X_amplitudes):

    df_decrements = pd.DataFrame()

    for i in range(len(X_amplitudes) - 2):
        s1 = X_amplitudes.iloc[i]
        s2 = X_amplitudes.iloc[i + 2]

        decrement = s1 / s2
        decrement.name = s1.name
        df_decrements = df_decrements.append(decrement)

    df_decrements['zeta_n'] = 1 / (2 * np.pi) * np.log(df_decrements['phi'])

    X_amplitudes_new = X_amplitudes.copy()
    X_amplitudes_new = X_amplitudes_new.iloc[0:-1].copy()
    X_amplitudes_new['zeta_n'] = df_decrements['zeta_n'].copy()
    X_amplitudes_new['B_n'] = 2*X_amplitudes_new['zeta_n']  # [Nm*s]

    return X_amplitudes_new


def fft_omega0(frequencies, dft):

    index = np.argmax(dft)
    natural_frequency = frequencies[index]
    omega0 = 2 * np.pi * natural_frequency
    return omega0

def fft(series):
    """
    FFT of a series
    Parameters
    ----------
    series

    Returns
    -------

    """

    signal = series.values
    time = series.index

    dt = np.mean(np.diff(time))
    #n = 11*len(time)
    n = 50000
    frequencies = np.fft.rfftfreq(n=n, d=dt) # [Hz]

    dft = np.abs(np.fft.rfft(signal, n=n))

    return frequencies, dft


def linearized_matrix(df_rolldecay, df_ikeda, phi_as = np.deg2rad(np.linspace(1,10,10)), g=9.81, rho=1000,
                      do_hatify=True, suffixes=('','_ikeda')):
    """
    Calculate B_e equivalent linearized damping for a range of roll amplitudes for both model tests and simplified ikeda.

    Parameters
    ----------
    df_rolldecay
    df_ikeda
    phi_as

    Returns
    -------

    """


    df = pd.DataFrame()

    for phi_a in phi_as:
        df_ = linearize(phi_a=phi_a, df_rolldecay=df_rolldecay, df_ikeda=df_ikeda, g=g, rho=rho, do_hatify=do_hatify,
                        suffixes=suffixes)
        df_['phi_a']=phi_a
        df =df.append(df_, ignore_index=True)

    return df


def linearize_si(phi_a, df_ikeda, components = ['B_44', 'B_F', 'B_W', 'B_E', 'B_BK', 'B_L'], do_hatify=True):
    """
    Calculate the equivalent linearized damping B_e

    Parameters
    ----------
    phi_a
    df_ikeda
    g
    rho

    Returns
    -------

    """

    df_ikeda = df_ikeda.copy()

    for component in components:
        new_key = '%s_e' % component
        B1_key = '%s_1' % component
        B2_key = '%s_2' % component

        df_ikeda[new_key] = lambdas.B_e_lambda(B_1=df_ikeda[B1_key],
                                               B_2=df_ikeda[B2_key],
                                               omega0=df_ikeda['omega0'],
                                               phi_a=phi_a)

    if do_hatify:
        df_ikeda['B_e'] = df_ikeda['B_44_e']
    else:
        df_ikeda['B_e_hat'] = df_ikeda['B_44_hat_e']

    return df_ikeda

def hatify(df_ikeda, g=9.81, rho=1000, components = ['B','B_44', 'B_F', 'B_W', 'B_E', 'B_BK', 'B_L']):

    df_ikeda=df_ikeda.copy()
    new_keys = ['%s_e' % key for key in components]
    new_hat_keys = ['%s_e_hat' % key for key in components]

    Disp = np.tile(df_ikeda['Disp'],[len(components),1]).T
    beam = np.tile(df_ikeda['beam'],[len(components),1]).T

    df_ikeda[new_hat_keys] = lambdas.B_e_hat_lambda(B_e=df_ikeda[new_keys],
                                                   Disp=Disp,
                                                   beam=beam,
                                                   g=g, rho=rho)

    df_ikeda['B_e_hat'] = df_ikeda['B_44_e_hat']

    return df_ikeda


def linearize_model_test(phi_a, df_rolldecay, g=9.81, rho=1000):
    """
    Calculate the equivalent linearized damping B_e

    Parameters
    ----------
    phi_a
    df_rolldecay
    g
    rho

    Returns
    -------

    """

    df_rolldecay = df_rolldecay.copy()

    df_rolldecay['B_e'] = lambdas.B_e_lambda(B_1=df_rolldecay['B_1'],
                                             B_2=df_rolldecay['B_2'],
                                             omega0=df_rolldecay['omega0'],
                                             phi_a=phi_a)

    df_rolldecay['B_e_hat'] = lambdas.B_e_hat_lambda(B_e=df_rolldecay['B_e'],
                                                     Disp=df_rolldecay['Disp'],
                                                     beam=df_rolldecay['beam'],
                                                     g=g, rho=rho)

    return df_rolldecay

def linearize(phi_a:float, df_rolldecay:pd.DataFrame, df_ikeda:pd.DataFrame, g=9.81, rho=1000,
              components = ['B_44', 'B_F', 'B_W', 'B_E', 'B_BK', 'B_L'], do_hatify=True, suffixes=('','_ikeda')):

    if not do_hatify:
        components = ['%s_hat' % key for key in components]

    df_rolldecay = linearize_model_test(phi_a=phi_a, df_rolldecay=df_rolldecay, g=g, rho=rho)
    df_ikeda = linearize_si(phi_a=phi_a, df_ikeda=df_ikeda, components=components, do_hatify=do_hatify)

    if do_hatify:
        df_ikeda = hatify(df_ikeda=df_ikeda,g=g, rho=rho, components=components)


    df_compare = pd.merge(left=df_rolldecay, right=df_ikeda, how='inner', left_index=True, right_index=True,
                          suffixes=suffixes)
    return df_compare



import rolldecayestimators.lambdas as lambdas
from scipy.optimize import least_squares

def residual_cubic(x, y, phi_a, omega0):
    """
    Residual function for least square fit
    """
    B_1 = x[0]
    B_2 = x[1]
    B_3 = x[2]

    B_e_pred = lambdas.B_e_lambda_cubic(B_1=B_1, B_2=B_2, B_3=B_3, omega0=omega0, phi_a=phi_a)
    B_e_true = y
    error = B_e_true - B_e_pred
    return error

def fit_cubic(y, phi_a, omega0, B_1_0=0,B_2_0=0,B_3_0=0):

    ## Use least square fit of B_44 as a function of phi_a to determine B_1, B_2 and B_3:
    x0 = [B_1_0,
          B_2_0,
          B_3_0,
          ]
    kwargs = {
        'y': y,
        'omega0': omega0,
        'phi_a':phi_a,
    }

    result = least_squares(fun=residual_cubic, x0=x0, kwargs=kwargs, method='lm')
    assert result.success

    output = {
        'B_1':result.x[0],
        'B_2': result.x[1],
        'B_3': result.x[2],
    }

    return output

def residual_quadratic(x, y, phi_a, omega0):
    """
    Residual function for least square fit
    """
    B_1 = x[0]
    B_2 = x[1]

    B_e_pred = lambdas.B_e_lambda(B_1=B_1, B_2=B_2, omega0=omega0, phi_a=phi_a)
    B_e_true = y
    error = B_e_true - B_e_pred
    return error

def fit_quadratic(y, phi_a, omega0, B_1_0=0,B_2_0=0):

    ## Use least square fit of B_44 as a function of phi_a to determine B_1, B_2 and B_3:
    x0 = [B_1_0,
          B_2_0,
          ]
    kwargs = {
        'y': y,
        'omega0': omega0,
        'phi_a':phi_a,
    }

    result = least_squares(fun=residual_quadratic, x0=x0, kwargs=kwargs, method='lm')
    assert result.success

    output = {
        'B_1':result.x[0],
        'B_2': result.x[1],
    }

    return output


import pytest

import pandas as pd
import numpy as np
from numpy.testing import assert_almost_equal
from rolldecayestimators.ikeda_estimator import IkedaEstimator
import matplotlib.pyplot as plt

T0 = 20.0
omega0 = 2.0*np.pi/T0
d=0.0
zeta = 0.044

lpp=100
TA=5
TF=5
beam=10
BKL=0.8*lpp
BKB=0.7
A0=0.95
kg=10
Volume = lpp*beam*TA*0.75
gm=0.5

@pytest.fixture
def df_roll_decay():

    phi0 = np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)
    estimator = IkedaEstimator(lpp=lpp, TA=TA, TF=TF, beam=beam, BKL=BKL, BKB=BKB, A0=A0, kg=kg, Volume=Volume, gm=gm,
                               V=0)
    IkedaEstimator.parameters = {
        'omega0':omega0,
        'zeta':zeta,
        'd':d,
    }

    yield estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

def check(X, estimator, omega0, d, zeta, decimal=4):
    estimator.fit(X=X)
    #assert estimator.result['success']

    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

    score = estimator.score(X)
    assert score > 0.90  # Have reached 0.98 before...

    #assert_almost_equal(X['phi'].values, X_pred['phi'].values, decimal=decimal)
    #assert estimator.score(X) > 0.999
    #assert_almost_equal(estimator.parameters['omega0'], omega0, decimal=decimal)
    #assert_almost_equal(estimator.parameters['zeta'], zeta, decimal=decimal)
    #assert_almost_equal(estimator.parameters['d'], d, decimal=decimal)


def test_fit(df_roll_decay):

    direct_estimator = IkedaEstimator(lpp=lpp, TA=TA, TF=TF, beam=beam, BKL=BKL, BKB=BKB, A0=A0, kg=kg, Volume=Volume,
                                      gm=gm, V=0, verify_input=False)
    X = df_roll_decay
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta, decimal=2)






import pytest

from sklearn.utils.estimator_checks import check_estimator

from rolldecayestimators import DirectEstimator, CutTransformer

@pytest.mark.skip('Write a real test later...')
@pytest.mark.parametrize(
    "Estimator", [DirectEstimator, CutTransformer]
)
def test_all_estimators(Estimator):
    return check_estimator(Estimator)



import numpy as np
import rolldecayestimators.lambdas as lambdas

def eddy(rho, lpp, d, beam, OG, phi_a, w):
    """
    Calculate eddy damping for a barge shaped hull accoring to:

    Ikeda, Y., 1993. Roll Damping of a Sharp-Cornered Barge and Roll Control by New Type Stabilizer,
    in: Proceedings of the 3rd International Offshore and Polar Engineering Conference.
    Presented at the The International Society of Offshore and Polar Engineers, Singapore.

    Parameters
    ----------
    rho
        water density [kg/m3]
    lpp
        ship length [m]
    d
        draught [m]
    beam
        ship beam [m]
    OG
         distance from the still water level O to the roll axis G [m]
    phi_a
        roll amplitude [rad]
    w
        frequency of roll motion [rad/s]
    Returns
    -------
    B_E
    """

    H_0 = beam/(2*d)
    B_E = (2/np.pi)*rho*lpp*d**4*(H_0**2 + 1 - OG/d)*(H_0**2 + (1 - OG/d)**2)*phi_a*w
    return B_E


import sympy as sp
import sympy.physics.mechanics as me

import rolldecayestimators.special_symbol as ss
import sympy as sp
import sympy.physics.mechanics as me

t = ss.Symbol(name='t',description='time',unit='s')
I = ss.Symbol(name='I',description='total roll inertia',unit='kg*m**2')
B = ss.Symbol(name='B',description='total roll damping',unit='kg*m*/s')
rho = ss.Symbol(name='rho',description='water density',unit='kg/m**3')
g = ss.Symbol(name='g',description='acceleration of gravity',unit='m/s**2')
Disp = ss.Symbol(name='Disp',description='displacement',unit='m**3')
M_x = ss.Symbol(name='M_x',description='External roll moment',unit='Nm')
m = ss.Symbol(name='m',description='mass of ship',unit='kg')
GM = ss.Symbol(name='GM', description='metacentric height', unit='m')
dGM = ss.Symbol(name='dGM', description='metacentric height correction', unit='m/rad')
omega = ss.Symbol(name='omega', description='Angular velocity of external moment', unit='rad/s')
L_pp = ss.Symbol(name='L_pp',description='ship perpendicular length',unit='m')
beam = ss.Symbol(name='beam',description='ship beam',unit='m')
x_s = ss.Symbol(name='x_s',description='section x-coordinate',unit='m')
AP = ss.Symbol(name='AP',description='ship perpendiculars',unit='-')
FP = ss.Symbol(name='FP',description='ship perpendiculars',unit='-')

C_p = ss.Symbol(name='C_p',description='Prismatic coefficient',unit='-')
C_b = ss.Symbol(name='C_b',description='Block coefficient',unit='-')

I_RUD = ss.Symbol(name='I_RUD',description='Number of rudders',unit='-')
BK_L = ss.Symbol(name='BK_L',description='Bilge keel length',unit='m')
BK_B = ss.Symbol(name='BK_B',description='Bilge keel height',unit='m')
A_0 = ss.Symbol(name='A_0',description='Mid ship area coefficient',unit='-')
ship_type_id = ss.Symbol(name='ship_type',description='Ship type',unit='-')
I_xx = ss.Symbol(name='I_xx',description='Roll intertia',unit='kg*m**2')
K_xx = ss.Symbol(name='K_xx',description='Nondimensional roll radius of gyration',unit='-')
R_h = ss.Symbol(name='R_h',description='Rudder height',unit='m')
A_R = ss.Symbol(name='A_R',description='Rudder area',unit='m**2')
TWIN = ss.Symbol(name='twin',description='Twin skrew',unit='True/False')
kg = ss.Symbol(name='kg',description='Keel to g',unit='m')
C_W = ss.Symbol(name='C_W',description='Water area coefficient',unit='-')
T_F = ss.Symbol(name='T_F',description='Draught forward',unit='m')
T_A = ss.Symbol(name='T_A',description='Draught aft',unit='m')
T = ss.Symbol(name='T',description='Mean draught',unit='m')
V = ss.Symbol(name='V',description='Ship speed',unit='m/s')
OG = ss.Symbol(name='OG',description='Distance into water from still water to centre of gravity',unit='m')

# Sections:

B_E0s = ss.Symbol(name="B'_E0", description='Zero speed sectional eddy damping', unit='Nm*s/(m)')
T_s = ss.Symbol(name='T_s',description='Section draught',unit='m')
B_s = ss.Symbol(name='B_s',description='Section beam',unit='m')
sigma = ss.Symbol(name='sigma',description='Section area coefficient',unit='-')

phi = me.dynamicsymbols('phi')  # Roll angle
#phi = ss.Symbol(name='phi', description='Roll angle', unit='rad')  # Roll angle
phi_dot = phi.diff()
phi_dot_dot = phi_dot.diff()
phi_a = ss.Symbol(name='phi_a', description='roll amplitude', unit='rad')

zeta = ss.Symbol(name='zeta', description='Damping coefficient', unit = '-') # Linear roll damping coefficeint
omega0 = ss.Symbol(name='omega0',description='Natural angular velocity',unit='rad/s')  # Natural roll frequency

d = sp.Symbol('d')  # Nonlinear roll damping coefficient

A_44 = ss.Symbol(name='A_44', description='Total mass moment of inertia', unit='kg*m**2')

B_1 = ss.Symbol(name='B_1',description='Linear damping coefficient',unit='Nm*s')  # Natural roll frequency
B_2 = ss.Symbol(name='B_2',description='Quadratic damping coefficient',unit='Nm*s**2')  # Natural roll frequency
B_3 = ss.Symbol(name='B_3',description='Cubic damping coefficient',unit='Nm*s**3')  # Natural roll frequency

C = ss.Symbol(name='C', description='General stiffness coefficient', unit=r'Nm/rad')  # Introducing a helper coefficient C

C_1 = ss.Symbol(name='C_1', description='Linear stiffness coefficient', unit=r'Nm')
C_3 = ss.Symbol(name='C_3',description='Stiffness coefficient', unit=r'Nm')
C_5 = ss.Symbol(name='C_5',description='Stiffness coefficient', unit=r'Nm')


B_e = ss.Symbol(name='B_e', description='Equivalen linearized damping', unit='Nm*s')
B_44_hat = ss.Symbol(name='B_44_hat', description='Nondimensional damping', unit='-')
B_e_hat = ss.Symbol(name='B_e_hat', description='Nondimensional damping', unit='-')
B_e_hat_0 = ss.Symbol(name='B_e_hat_0', description='Nondimensional damping', unit='-')
B_e_factor = ss.Symbol(name='B_e_factor', description='Nondimensional damping', unit='-')

B_W_e_hat = ss.Symbol(name='B_W_e_hat', description='Nondimensional damping', unit='-')
B_F_e_hat = ss.Symbol(name='B_F_e_hat', description='Nondimensional damping', unit='-')
B_BK_e_hat = ss.Symbol(name='B_BK_e_hat', description='Nondimensional damping', unit='-')
B_E_e_hat = ss.Symbol(name='B_E_e_hat', description='Nondimensional damping', unit='-')
B_L_e_hat = ss.Symbol(name='B_L_e_hat', description='Nondimensional damping', unit='-')


B_1_hat = ss.Symbol(name='B_1_hat', description='Nondimensional damping', unit='-')
B_2_hat = ss.Symbol(name='B_2_hat', description='Nondimensional damping', unit='-')
omega_hat = ss.Symbol(name='omega_hat', description='Nondimensional roll frequency', unit='-')
omega0_hat = ss.Symbol(name='omega0_hat', description='Nondimensional roll frequency', unit='-')

B_1_hat0 = ss.Symbol(name='B_1_hat0', description='Nondimensional damping at zero speed', unit='-')

B_44_ = ss.Symbol(name='B_44', description='Total roll damping at a certain roll amplitude', unit='Nm*s')
B_F = ss.Symbol(name='B_F', description='Friction roll damping', unit='Nm*s')
B_W = ss.Symbol(name='B_W', description='Wave roll damping', unit='Nm*s')
B_E = ss.Symbol(name='B_E', description='Eddy roll damping', unit='Nm*s')
B_E0 = ss.Symbol(name='B_E0', description='Zero speed eddy damping ', unit='Nm*s')
B_BK = ss.Symbol(name='B_{BK}', description='Bilge keel roll damping', unit='Nm*s')
B_L = ss.Symbol(name='B_L', description='Hull lift roll damping', unit='Nm*s')




B_44_HAT =  ss.Symbol(name='B_44_HAT', description='Total roll damping at a certain roll amplitude', unit='-')
B_F_HAT = B_F_hat = ss.Symbol(name='B_F_HAT', description='Friction roll damping', unit='-')
B_W_HAT = B_W_hat = ss.Symbol(name='B_W_HAT', description='Wave roll damping', unit='-')
B_E_HAT = B_E_hat = ss.Symbol(name='B_E_HAT', description='Eddy roll damping', unit='-')
B_E0_hat = ss.Symbol(name='B_E0_HAT', description='Eddy roll damping at zero speed', unit='-')
B_BK_HAT = B_BK_hat = ss.Symbol(name='B_BK_HAT', description='Bilge keel roll damping', unit='-')
B_L_HAT = B_L_hat = ss.Symbol(name='B_L_HAT', description='Hull lift roll damping', unit='-')

## Functions:
GZ = sp.Function('GZ')(phi)
B_44 = sp.Function('B_{44}')(phi_dot)
C_44 = sp.Function('C_{44}')(phi)
M_44 = sp.Function('M_{44}')(omega*t)

## Analytical
y = me.dynamicsymbols('y')
y0 = me.dynamicsymbols('y0')
y0_dot = y0.diff()
y0_dotdot = y0_dot.diff()
D = sp.symbols('D')

phi_0 = me.dynamicsymbols('phi_0')
phi_0_dot = phi_0.diff()
phi_0_dotdot = phi_0_dot.diff()

ikeda_simplified = sp.Function('f')(L_pp,beam,C_b,A_0,
                                  OG,phi_a,BK_L,BK_B,omega,
                                  T, V)

"""
Ikeda, Y.,
1978. On eddy making component of roll damping force on naked hull. University of Osaka Prefacture,
Department of Naval Architecture, Japan, Report No. 00403,
Published in: Journal of Society of Naval Architects of Japan, Volume 142.
"""
C_P =  ss.Symbol(name='C_P', description='Pressure difference coefficient', unit='-')
C_r =  ss.Symbol(name='C_r', description='Eddy damping coefficient', unit='-')
r_max = ss.Symbol(name='r_max', description='Maximum distance from roll axis to hull', unit='m')
P_m = ss.Symbol(name='P_m', description='Pressure difference', unit='N/m**2')
R_b = ss.Symbol(name='R_b', description='Bilge radius', unit='m')
f_1 = ss.Symbol(name='f_1', description='Difference of flow factor', unit='-')
f_2 = ss.Symbol(name='f_2', description='Modification factor', unit='-')
H_0 = ss.Symbol(name='H_0', description='Half beam-draft ratio', unit='-')
B_E_star_hat = ss.Symbol(name='B_E_star_hat', description='Only nonlinear nondimensional eddy damping', unit='-')
B_F_star_hat = ss.Symbol(name='B_F_star_hat', description='Only nonlinear nondimensional friction damping', unit='-')
B_W_star_hat = ss.Symbol(name='B_W_star_hat', description='Only nonlinear nondimensional wave damping', unit='-')
B_star_hat = ss.Symbol(name='B_star_hat', description='Only nonlinear nondimensional damping', unit='-')


## Lewis
a_1 = ss.Symbol(name='a_1', description='Lewis section coefficient', unit='-')
a_3 = ss.Symbol(name='a_3', description='Lewis section coefficient', unit='-')
D_1= ss.Symbol(name='D_1', description='Lewis section coefficient', unit='-')


import pytest
import pandas as pd
import numpy as np
from rolldecayestimators import ikeda_naked
from rolldecayestimators import ikeda_speed

from rolldecayestimators.ikeda_speed import calculate_sectional_lewis
from rolldecayestimators import lambdas
from numpy.testing import assert_almost_equal

@pytest.mark.skip('This one does not work yet')
def test_eddy_sections_1():

    B_s = np.array([0.28])
    T_s = 0.112
    sigma = 1.0
    R_b = 0.0
    g = 9.81
    OG = 0.0
    L = 0.8
    volume = 0.02509
    rho=1000.0

    S_s = sigma * B_s * T_s
    a, a_1, a_3, sigma_s, H0 = calculate_sectional_lewis(B_s=B_s, T_s=T_s, S_s=S_s)
    assert H0==1.25

    w_hat = 1.0
    phi_a = 0.175
    w = lambdas.omega_from_hat(beam=B_s, g=g, omega_hat=w_hat)

    #R_b = ikeda_eddy.calculate_R_b(beam=B_s, draught=T_s, H0=H0, sigma=sigma_s)

    B_E0_s = ikeda_speed.eddy_sections(bwl=B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, H0=H0, Ts=T_s, OG=OG, R=R_b, wE=w,
                                       fi_a=phi_a, ra=rho)

    B_E0 = B_E0_s * L

    Disp = volume
    B_E0_hat = lambdas.B_to_hat_lambda(B=B_E0, Disp=Disp, beam=B_s, g=g, rho=rho)
    B_E0_star_hat = B_E0_hat * 3 * np.pi / 8

    assert_almost_equal(B_E0_star_hat, 0.042)

@pytest.mark.skip('This one does not work yet')
def test_eddy_sections_2():

    B_s = 0.28
    T_s = 0.112
    sigma = 1.0
    R_b = 0.0
    g = 9.81
    OG = 0.0
    L = 0.8
    volume = 0.02509
    rho=1000.0

    S_s = sigma * B_s * T_s
    a, a_1, a_3, sigma_s, H0 = calculate_sectional_lewis(B_s=B_s, T_s=T_s, S_s=S_s)
    assert H0==1.25

    w_hat = 1.0
    phi_a = 0.175
    w = lambdas.omega_from_hat(beam=B_s, g=g, omega_hat=w_hat)

    B_E0_s = ikeda_naked.eddy_sections(bwl=B_s, a_1=a_1, a_3=a_3, sigma=sigma_s, H0=H0, Ts=T_s,
                  OG=OG, R=R_b, wE=w, fi_a=phi_a, ra=rho)

    B_E0 = B_E0_s * L

    Disp = volume
    B_E0_hat = lambdas.B_to_hat_lambda(B=B_E0, Disp=Disp, beam=B_s, g=g, rho=rho)
    B_E0_star_hat = B_E0_hat * 3 * np.pi / 8

    assert_almost_equal(B_E0_star_hat, 0.042)

def test_calculate_Cr():

    df_kvlcc2 = pd.Series({
        'area' :    917.895066,
        'x' :   61.597568,
        'd':    20.800000,
        'B':    56.159232,
        'R':    34.146143,
        'sigma':    0.785794,
        'OG/d':    0.000000,
        'a_1':    0.148893,
        'a_3': - 0.000246,
        'H0':    1.349982,
    })

    OG = df_kvlcc2['OG/d']*df_kvlcc2.d
    ra = 1000
    df_kvlcc2 = pd.DataFrame(df_kvlcc2).transpose()

    C_r = ikeda_naked.calculate_C_r(bwl=df_kvlcc2.B,
                          a_1=df_kvlcc2.a_1, a_3=df_kvlcc2.a_3, sigma=df_kvlcc2.sigma, H0=df_kvlcc2.H0, d=df_kvlcc2.d,
                          OG=OG,
                          R=df_kvlcc2.R, ra=ra)


import sympy as sp
from rolldecayestimators.symbols import *

# Some of the symbols can be nicer displayed using this:
nicer_LaTeX = [
        (B_E0_hat, sp.symbols('\hat{B}_{E0}')),
        (B_E_star_hat, sp.symbols('\hat{B^*}_{E}')),
        (B_W_star_hat, sp.symbols('\hat{B^*}_{W}')),
        (B_F_star_hat, sp.symbols('\hat{B^*}_{F}')),
        (B_star_hat, sp.symbols('\hat{B^*}')),
        (omega_hat, sp.symbols('\hat{\omega}')),
                
    ]

# General roll motion equation according to Himeno:
lhs = A_44*phi_dot_dot + B_44 + C_44
rhs = M_44
roll_equation_himeno = sp.Eq(lhs=lhs, rhs=rhs)

# No external forces (during roll decay)
roll_decay_equation_general_himeno = roll_equation_himeno.subs(M_44,0)


restoring_equation = sp.Eq(C_44,m*g*GZ)
restoring_equation_linear = sp.Eq(C_44,m*g*GM*phi)
restoring_equation_quadratic = sp.Eq(C_44, C_1 * phi + C_3 * phi ** 3)
restoring_equation_cubic = sp.Eq(C_44,C_1*phi + C_3*phi*sp.Abs(phi) + C_5*phi**3)

## Cubic model:
b44_cubic_equation = sp.Eq(B_44, B_1 * phi_dot + B_2 * phi_dot * sp.Abs(phi_dot) + B_3 * phi_dot ** 3)
restoring_equation_cubic = sp.Eq(C_44, C_1 * phi + C_3 * phi ** 3 + C_5 * phi ** 5)

subs = [
    (B_44, sp.solve(b44_cubic_equation, B_44)[0]),
    (C_44, sp.solve(restoring_equation_cubic, C_44)[0])
]
roll_decay_equation_himeno_quadratic = roll_decay_equation_general_himeno.subs(subs)


## Quadratic model:
b44_quadratic_equation = sp.Eq(B_44, B_1 * phi_dot + B_2 * phi_dot * sp.Abs(phi_dot))
restoring_equation_quadratic = sp.Eq(C_44, C_1 * phi + C_3 * phi ** 3)

subs = [
    (B_44, sp.solve(b44_quadratic_equation, B_44)[0]),
    (C_44, sp.solve(restoring_equation_quadratic, C_44)[0])
]
roll_decay_equation_himeno_quadratic = roll_decay_equation_general_himeno.subs(subs)

## Linear model:
b44_linear_equation = sp.Eq(B_44, B_1 * phi_dot)
restoring_linear_quadratic = sp.Eq(C_44, C_1 * phi)

subs = [
    (B_44, sp.solve(b44_linear_equation, B_44)[0]),
    (C_44, sp.solve(restoring_linear_quadratic, C_44)[0])
]
roll_decay_equation_himeno_linear = roll_decay_equation_general_himeno.subs(subs)

C_equation = sp.Eq(C,C_44/phi)
C_equation_linear = C_equation.subs(C_44,sp.solve(restoring_equation_linear,C_44)[0])
C_equation_cubic = C_equation.subs(C_44,sp.solve(restoring_equation_cubic,C_44)[0])
C_equation_quadratic = C_equation.subs(C_44,sp.solve(restoring_equation_quadratic,C_44)[0])

roll_decay_equation_himeno_quadratic =  roll_decay_equation_general_himeno.subs(B_44,
                                                        sp.solve(b44_quadratic_equation,B_44)[0]).subs(C,
                                                            sp.solve(C_equation_quadratic,C)[0])

subs = [
    (B_44, sp.solve(b44_quadratic_equation, B_44)[0]),
    (C_44, sp.solve(restoring_linear_quadratic, C_44)[0])
]
roll_decay_equation_himeno_quadratic_b = roll_decay_equation_general_himeno.subs(subs)


roll_decay_equation_himeno_quadratic_c = roll_decay_equation_himeno_quadratic.subs(C_44,sp.solve(C_equation, C_44)[0])
zeta_equation = sp.Eq(2*zeta*omega0,B_1/A_44)
d_equation = sp.Eq(d,B_2/A_44)
omega0_equation = sp.Eq(omega0,sp.sqrt(C/A_44))

eq = sp.Eq(roll_decay_equation_himeno_quadratic_c.lhs/A_44,0)  # helper equation

subs = [
    (B_1, sp.solve(zeta_equation, B_1)[0]),
    (B_2, sp.solve(d_equation, B_2)[0]),
    (C / A_44, sp.solve(omega0_equation, C / A_44)[0])

]

roll_decay_equation_quadratic = sp.Eq(sp.expand(eq.lhs).subs(subs), 0)
roll_decay_equation_quadratic = sp.factor(roll_decay_equation_quadratic, phi_dot)
roll_decay_equation_linear = roll_decay_equation_quadratic.subs(d,0)

omega0_equation_linear = omega0_equation.subs(C,sp.solve(C_equation_linear,C)[0])
A44 = sp.solve(omega0_equation_linear, A_44)[0]
zeta_B1_equation = zeta_equation.subs(A_44,A44)
d_B2_equation = d_equation.subs(A_44,A44)

## Cubic model:
subs = [
    (B_44,sp.solve(b44_cubic_equation,B_44)[0]),
    (C_44,sp.solve(restoring_equation_cubic,C_44)[0])
]
roll_decay_equation_cubic = roll_decay_equation_general_himeno.subs(subs)
# But this equation does not have a unique solution, so we devide all witht the interia A_44:

normalize_symbols = [B_1, B_2, B_3, C_1, C_3, C_5]
normalize_equations = {}
new_symbols = {}
subs_normalize = []
for symbol in normalize_symbols:
    
    if 'C' in symbol.name:
        description = 'Stiffness help coefficients'
    elif 'B' in symbol.name:
        description = 'Damping help coefficients'
    else:
        description = ' '
    
    new_symbol = ss.Symbol(name='%sA' % symbol.name, description=description, unit=' ')
    new_symbol = sp.Symbol('%sA' % symbol.name)
    new_symbols[symbol] = new_symbol
    eq = sp.Eq(new_symbol,symbol/A_44)

    normalize_equations[symbol]=eq

    subs_normalize.append((symbol, sp.solve(eq, symbol)[0]))

lhs = (roll_decay_equation_cubic.lhs/A_44).subs(subs_normalize).simplify()
roll_decay_equation_cubic_A = sp.Eq(lhs=lhs,rhs=0)


## Equivalt linearized damping:
B_e_equation = sp.Eq(B_e,B_1+8/(3*sp.pi)*omega0*phi_a*B_2)
B_e_equation_cubic = sp.Eq(B_e,B_1+8/(3*sp.pi)*omega0*phi_a*B_2 + 3/4*omega0**2*phi_a**2*B_3)


A_44_eq = sp.Eq(A_44, A44)
eqs = [
    A_44_eq,
    C_equation_linear,

]

omega0_eq = sp.Eq(omega0,sp.solve(eqs, omega0, GM)[1][0])
omega0_eq = omega0_eq.subs(C,C_1)

## Nondimensional damping Himeno:
lhs = B_44_hat
rhs = B_44/(rho*Disp*beam**2)*sp.sqrt(beam/(2*g))
B44_equation = sp.Eq(lhs, rhs)
omega0_equation_linear = omega0_equation.subs(C,sp.solve(C_equation_linear,C)[0])

omega_hat_equation = sp.Eq(omega_hat,omega*sp.sqrt(beam/(2*g)))
B44_hat_equation = sp.Eq(B_44_hat, B_44/(rho*Disp*beam**2)*sp.sqrt(beam/(2*g)))
B_1_hat_equation = sp.Eq(B_1_hat, B_1/(rho*Disp*beam**2)*sp.sqrt(beam/(2*g)))
B_e_hat_equation = sp.Eq(B_e_hat, B_e/(rho*Disp*beam**2)*sp.sqrt(beam/(2*g)))

B_2_hat_equation = sp.Eq(B_2_hat, B_2/(rho*Disp*beam**2)*sp.sqrt(beam/(2*g))**(0))

B44_hat_equation_quadratic = B44_hat_equation.subs(B_44,sp.solve(b44_quadratic_equation,B_44)[0])
omega0_hat_equation = omega_hat_equation.subs(omega,omega0)


## Analytical
diff_eq = sp.Eq(y.diff().diff() + 2 * zeta * omega0 * y.diff() + omega0 ** 2 * y, 0)
equation_D = sp.Eq(D, sp.sqrt(1 - zeta ** 2))

lhs = y
rhs = sp.exp(-zeta * omega0 * t) * (y0 * sp.cos(omega0 * D * t) + (y0_dot / (omega0 * D) + zeta * y0 / D) * sp.sin(omega0 * D * t))

analytical_solution_general = sp.Eq(lhs,rhs)

subs = [
    (y,phi),
    (y0, phi_0),
    (y0_dot, phi_0_dot),
    (y0_dotdot, phi_0_dotdot),
    (D,sp.solve(equation_D,D)[0]),
]

analytical_solution = analytical_solution_general.subs(subs)
analytical_phi1d = sp.Eq(phi_dot,sp.simplify(analytical_solution.rhs.diff(t)))
analytical_phi2d = sp.Eq(phi_dot_dot,sp.simplify(analytical_phi1d.rhs.diff(t)))

rhs = analytical_solution.rhs.args[1]*phi_0
lhs = phi_a
extinction_equation = sp.Eq(lhs,rhs)

xeta_equation = sp.Eq(zeta,
      sp.solve(extinction_equation,zeta)[0])

B_1_zeta_eq = sp.Eq(B_1, 2*zeta*omega0*A_44)
B_1_zeta_eq

### Simplified Ikeda
simplified_ikeda_equation = sp.Eq((B_F,
      B_W,
      B_E,
      B_BK,
       B_L,
      ),ikeda_simplified)


### Regression
regression_factor_equation = sp.Eq(B_e_hat,B_e_hat_0*B_e_factor)


import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_is_fitted

from rolldecayestimators import DirectEstimator, measure as measure


class NorwegianEstimator(DirectEstimator):

    def __init__(self, maxfev = 4000, bounds=None, ftol=10**-10):
        super().__init__(maxfev=maxfev,bounds=bounds, ftol=ftol)
        self.phi_key = 'phi'  # Roll angle [rad]
        self.phi1d_key = 'phi1d'  # Roll velocity [rad/s]
        self.phi2d_key = 'phi2d'  # Roll acceleration [rad/s2]

    def fit(self, X, y=None):
        """A reference implementation of a fitting function.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : Dummy not used.

        Returns
        -------
        self : object
            Returns self.
        """
        #X, y = check_X_y(X, y, accept_sparse=True)
        self.X = X.copy()
        self.is_fitted_ = True
        # `fit` should always return `self`

        self.calculate_amplitudes_and_damping()

        # Fitting part:
        self.linear_regression = LinearRegression()
        self.X_amplitudes['x'] = 2 / (3 * np.pi) * self.X_amplitudes['phi']

        self.linear_regression.fit(X=self.X_amplitudes[['x']], y=self.X_amplitudes['zeta_n'])

        omega0 = self.omega0
        self.parameters = {
            'd' : self.linear_regression.coef_[0],
            'zeta' : self.linear_regression.intercept_,
            'omega0' : omega0,
        }

        return self




import pytest

import pandas as pd
import numpy as np
from numpy.testing import assert_almost_equal
#from rolldecayestimators.simulation import simulate
from rolldecayestimators.direct_estimator import DirectEstimator
import matplotlib.pyplot as plt

@pytest.fixture
def d():
    yield 0.076

@pytest.fixture
def omega0():
    T0 = 20.0
    yield 2.0*np.pi/T0

@pytest.fixture
def zeta():
    yield 0.044

@pytest.fixture
def df_roll_decay(omega0, d, zeta):

    phi0 = np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)
    estimator = DirectEstimator.load(omega0=omega0, d=d, zeta=zeta)

    yield estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

@pytest.fixture
def df_roll_decay_negative(omega0, d, zeta):

    phi0 = -np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)
    estimator = DirectEstimator.load(omega0=omega0, d=d, zeta=zeta)
    yield estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

def check(X, estimator, omega0, d, zeta, decimal=4):
    estimator.fit(X=X)
    assert estimator.result['success']
    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

    assert_almost_equal(X['phi'].values, X_pred['phi'].values, decimal=decimal)
    assert estimator.score(X) > 0.999
    assert_almost_equal(estimator.parameters['omega0'], omega0, decimal=decimal)
    assert_almost_equal(estimator.parameters['zeta'], zeta, decimal=decimal)
    assert_almost_equal(estimator.parameters['d'], d, decimal=decimal)


def test_fit_simualtion_derivation(df_roll_decay, omega0, d, zeta):

    direct_estimator = DirectEstimator(fit_method='derivation', omega_regression=False)
    X = df_roll_decay
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta, decimal=2)  # Bad accuracy! 


def test_fit_simualtion_derivation_omega(df_roll_decay,omega0, d, zeta):

    direct_estimator = DirectEstimator(fit_method='derivation', omega_regression=True)

    X = df_roll_decay
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta, decimal=2)  # Bad accuracy!

def test_fit_simualtion_integration(df_roll_decay,omega0, d, zeta):

    direct_estimator = DirectEstimator(fit_method='integration', omega_regression=False)

    X = df_roll_decay
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta)

def test_fit_simualtion_integration_omega(df_roll_decay,omega0, d, zeta):

    direct_estimator = DirectEstimator(fit_method='integration', omega_regression=True)

    X = df_roll_decay
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta)

def test_fit_simualtion_integration_omega_negative(df_roll_decay_negative,omega0, d, zeta):

    direct_estimator = DirectEstimator(fit_method='integration', omega_regression=True)

    X = df_roll_decay_negative
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, omega0=omega0, d=d, zeta=zeta)





import numpy as np
import matplotlib.pyplot as plt
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.linear_model import LinearRegression
import pandas as pd

import rolldecayestimators.filters
import rolldecayestimators.measure as measure
from sklearn.metrics import r2_score

class CutTransformer(BaseEstimator, TransformerMixin):
    """ Rolldecay transformer that cut time series from roll decay test for estimator.

    Parameters
    ----------
    phi_max : float, default=np.deg2rad(90)
        Start cutting value is below this value [rad]

    phi_min : float, default=0
        Stop cutting value is when below this value [rad]

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.
    """
    def __init__(self, phi_max=np.deg2rad(90), phi_min=0, phi1d_start_tolerance=0.005):
        self.phi_max = phi_max  # Maximum Roll angle [rad]
        self.phi_min = phi_min  # Minimum Roll angle [rad]
        self.phi_key = 'phi'  # Roll angle [rad]
        self.remove_end_samples = 200  # Remove this many samples from end (funky stuff may happen during end of tests)
        self.phi1d_start_tolerance = phi1d_start_tolerance

    def fit(self, X, y=None):
        """Do the cut

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        #X = check_array(X, accept_sparse=True)

        self.n_features_ = X.shape[1]

        phi = X[self.phi_key]
        if (self.phi_max < phi.abs().min()):
            raise ValueError('"phi_max" is too small')

        if (self.phi_min > phi.abs().max()):
            raise ValueError('"phi_min" is too large')

        if not isinstance(self.remove_end_samples,int):
            raise ValueError('"remove_end_samples" should be integer')

        if self.remove_end_samples<1:
            raise ValueError('"remove_end_samples" > 1')

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : {array-like, sparse-matrix}, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_features)
            The array containing the element-wise square roots of the values
            in ``X``.
        """
        # Check is fit had been called
        check_is_fitted(self, 'n_features_')

        # Input validation
        #X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        #if X.shape[1] != self.n_features_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')

        #Remove initial part (by removing first to maximums):
        phi = X[self.phi_key]
        index = phi.abs().idxmax()
        X_cut = X.loc[index:].copy()

        if (len(X_cut) > 10*self.remove_end_samples):
            X_cut = X_cut.iloc[0:-self.remove_end_samples]

        phi = X_cut[self.phi_key]
        phi_max_sign = np.sign(phi.loc[index])
        if phi_max_sign == 1:
            index2 = phi.idxmin()
        else:
            index2 = phi.idxmax()

        X_cut = X_cut.loc[index2:].copy()

        X_interpolated = measure.sample_increase(X=X_cut, increase=5)
        X_zerocrossings = measure.get_peaks(X=X_interpolated)
        mask = X_interpolated.index >= X_zerocrossings.index[0]
        X_interpolated = X_interpolated.loc[mask]

        # Remove some large angles at start
        mask = X_zerocrossings['phi'].abs() < self.phi_max
        X_zerocrossings2 = X_zerocrossings.loc[mask].copy()
        if len(X_zerocrossings2) > 0:
            mask2 = X_interpolated.index > X_zerocrossings2.index[0]
            X_interpolated = X_interpolated.loc[mask2]


        # Remove some small angles at end
        mask = X_zerocrossings2['phi'].abs() < self.phi_min
        X_zerocrossings3 = X_zerocrossings2.loc[mask].copy()
        if len(X_zerocrossings3) > 0:
            mask3 = X_interpolated.index < X_zerocrossings3.index[0]
            X_interpolated = X_interpolated.loc[mask3]

        if 'phi1d' in X_cut:
            phi1d_start = np.abs(X_interpolated.iloc[0]['phi1d'])
        
            if phi1d_start > self.phi1d_start_tolerance:
                raise ValueError('Start phi1d exceeds phi1d_start_tolerance (%f > %f)' % (phi1d_start, self.phi1d_start_tolerance) )

        mask = ((X_cut.index >= X_interpolated.index[0]) & (X_cut.index <= X_interpolated.index[-1]))
        X_cut=X_cut.loc[mask].copy()

        return X_cut

class LowpassFilterDerivatorTransformer(BaseEstimator, TransformerMixin):
    """ Rolldecay transformer that lowpass filters the roll signal for estimator.

    Parameters
    ----------
    phi_max : float, default=np.deg2rad(90)
        Start cutting value is below this value [rad]

    phi_min : float, default=0
        Stop cutting value is when below this value [rad]

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.
    """
    def __init__(self, cutoff=0.5, order=5, minimum_score=0.999):
        self.cutoff = cutoff
        self.order = order
        self.phi_key = 'phi'  # Roll angle [rad]
        self.phi_filtered_key = 'phi_filtered'  # Filtered roll angle [rad]
        self.phi1d_key = 'phi1d'  # Roll velocity [rad/s]
        self.phi2d_key = 'phi2d'  # Roll acceleration [rad/s2]
        self.minimum_score = minimum_score

    def fit(self, X, y=None):
        """Do the cut

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        #X = check_array(X, accept_sparse=True)

        self.n_features_ = X.shape[1]

        assert self.score(X=X) > self.minimum_score

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : {array-like, sparse-matrix}, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_features)
            The array containing the element-wise square roots of the values
            in ``X``.
        """
        # Check is fit had been called
        check_is_fitted(self, 'n_features_')

        # Input validation
        #X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        #if X.shape[1] != self.n_features_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')

        # Lowpass filter the signal:
        self.X = X.copy()
        self.X_filter = X.copy()
        ts = np.mean(np.diff(self.X_filter.index))
        fs = 1 / ts
        self.X_filter[self.phi_filtered_key] = rolldecayestimators.filters.lowpass_filter(data=self.X_filter['phi'],
                                                                                     cutoff=self.cutoff, fs=fs,
                                                                                     order=self.order)

        self.X_filter = self.add_derivatives(X=self.X_filter)

        return self.X_filter

    def plot_filtering(self):

        fig, axes = plt.subplots(nrows=3)

        ax = axes[0]
        self.X.plot(y='phi', ax=ax)
        self.X_filter.plot(y='phi_filtered', ax=ax, style='--')
        ax.legend();

        ax = axes[1]
        self.X_filter.plot(y='phi1d', ax=ax, style='--')
        ax.legend();

        ax = axes[2]
        self.X_filter.plot(y='phi2d', ax=ax, style='--')
        ax.legend();

    def add_derivatives(self, X):
        # Add accelerations:
        assert self.phi_key in X
        X = X.copy()
        X[self.phi1d_key] = np.gradient(X[self.phi_filtered_key].values, X.index.values)
        X[self.phi2d_key] = np.gradient(X[self.phi1d_key].values, X.index.values)
        return X

    def score(self, X, y=None, sample_weight=None):
        """
        Return the coefficient of determination R_b^2 of the prediction.

        The coefficient R_b^2 is defined as (1 - u/v), where u is the residual sum of squares
        ((y_true - y_pred) ** 2).sum() and v is the total sum of squares ((y_true - y_true.mean()) ** 2).sum().
        The best possible score is 1.0 and it can be negative (because the model can be arbitrarily worse).
        A constant model that always predicts the expected value of y, disregarding the input features,
        would get a R_b^2 score of 0.0.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples. For some estimators this may be a precomputed kernel matrix or a list of generic
            objects instead, shape = (n_samples, n_samples_fitted), where n_samples_fitted is the number of samples
            used in the fitting for the estimator.

        y : Dummy not used

        sample_weight : Dummy

        Returns
        -------
        score : float
            R_b^2 of self.predict(X) wrt. y.

        """
        X_filter = self.transform(X)
        y_true = X[self.phi_key]
        y_pred = X_filter[self.phi_filtered_key]

        return r2_score(y_true=y_true, y_pred=y_pred)

class ScaleFactorTransformer(BaseEstimator, TransformerMixin):
    """ Rolldecay to full scale using scale factor

    Parameters
    ----------
    phi_max : float, default=np.deg2rad(90)
        Start cutting value is below this value [rad]

    phi_min : float, default=0
        Stop cutting value is when below this value [rad]

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.
    """
    def __init__(self, scale_factor):
        self.scale_factor = scale_factor
        self.phi1d_key = 'phi1d'  # Roll velocity [rad/s]
        self.phi2d_key = 'phi2d'  # Roll acceleration [rad/s2]

    def fit(self, X, y=None):
        """Do the cut

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        #X = check_array(X, accept_sparse=True)

        self.n_features_ = X.shape[1]

        if pd.isnull(self.scale_factor):
            raise ValueError('Bad scale factor:%s' % self.scale_factor)

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : {array-like, sparse-matrix}, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_features)
            The array containing the element-wise square roots of the values
            in ``X``.
        """
        # Check is fit had been called
        check_is_fitted(self, 'n_features_')

        # Input validation
        #X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        #if X.shape[1] != self.n_features_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')


        X_scaled = X.copy()
        X_scaled.index*=np.sqrt(self.scale_factor)  # To full scale
        if self.phi1d_key in X:
            X_scaled[self.phi1d_key]/=np.sqrt(self.scale_factor)

        if self.phi2d_key in X:
            X_scaled[self.phi2d_key]/=self.scale_factor


        return X_scaled

class OffsetTransformer(BaseEstimator, TransformerMixin):
    """ Rolldecay remove offset in signal

    Parameters
    ----------
    phi_max : float, default=np.deg2rad(90)
        Start cutting value is below this value [rad]

    phi_min : float, default=0
        Stop cutting value is when below this value [rad]

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.
    """
    def __init__(self):
        self.phi1d_key = 'phi1d'  # Roll velocity [rad/s]
        self.phi2d_key = 'phi2d'  # Roll acceleration [rad/s2]

    def fit(self, X, y=None):
        """Do the cut

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        #X = check_array(X, accept_sparse=True)

        self.n_features_ = X.shape[1]

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : {array-like, sparse-matrix}, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_transformed : array, shape (n_samples, n_features)
            The array containing the element-wise square roots of the values
            in ``X``.
        """
        # Check is fit had been called
        check_is_fitted(self, 'n_features_')

        # Input validation
        #X = check_array(X, accept_sparse=True)

        # Check that the input is of the same shape as the one passed
        # during fit.
        #if X.shape[1] != self.n_features_:
        #    raise ValueError('Shape of input is different from what was seen'
        #                     'in `fit`')


        X_offset = X.copy()

        #X_interpolated = measure.sample_increase(X=X)
        X_zerocrossings = measure.get_peaks(X=X_offset)

        linear_regression = LinearRegression(fit_intercept=False)
        X_ = np.array([X_zerocrossings.index.values]).transpose()
        linear_regression.fit(X=X_, y=X_zerocrossings['phi'])
        X_zerocrossings['phi0'] = linear_regression.predict(X=X_)
        X_zerocrossings['phi_'] = X_zerocrossings['phi'] - X_zerocrossings['phi0']

        X_2 = np.array([X_offset.index.values]).transpose()
        X_offset['phi0'] = linear_regression.predict(X=X_2)
        X_offset['phi_offset'] = X_offset['phi'].copy()
        X_offset['phi'] = X_offset['phi'] - X_offset['phi0']

        self.X = X_offset

        return X_offset

    def plot_correction_line(self, ax=None):

        check_is_fitted(self, 'n_features_')

        if ax is None:
            fig,ax=plt.subplots()

        self.X_zerocrossings.plot(y='phi', style='ro', ax=ax, label='phi')
        self.X_zerocrossings.plot(y='phi0', style='b-', ax=ax, label='correction line')
        ax.plot([np.min(self.X_zerocrossings.index), np.max(self.X_zerocrossings.index)], [0, 0], 'g-')
        ax.legend()

    def plot(self, ax=None):

        check_is_fitted(self, 'n_features_')

        if ax is None:
            fig,ax=plt.subplots()

        self.X.plot(y='phi_offset', ax=ax, label='old')
        self.X.plot(y='phi', ax=ax, label='corrected')
        ax.plot([np.min(self.X.index), np.max(self.X.index)], [0, 0], 'g-')
        ax.legend()



import pytest
import numpy as np
from numpy.testing import assert_almost_equal
import matplotlib.pyplot as plt

from rolldecayestimators.estimator import RollDecay

T0 = 20
omega0 = 2 * np.pi / T0
zeta = 0.044

def simulator(estimator):
    phi0 = np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)

    estimator.parameters = {
         'omega0':omega0,
         'zeta': zeta,
    }

    return estimator.simulate(t=t, phi0 = phi0, phi1d0 = phi1d0,)

def simulator_negative(estimator):
    phi0 = -np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)

    estimator.parameters = {
        'omega0': omega0,
        'zeta': zeta,
    }

    return estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

def simulator_sparse(estimator):
    phi0 = np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 10)

    estimator.parameters = {
        'omega0': omega0,
        'zeta': zeta,
    }

    df = estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)
    return df

def simulator_noise(estimator):
    phi0 = np.deg2rad(2)
    phi1d0 = 0
    t = np.arange(0, 120, 0.1)
    estimator.parameters = {
        'omega0': omega0,
        'zeta': zeta,
    }

    df = estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)
    np.random.seed(0)
    std = np.deg2rad(0.1)
    noise = np.random.normal(loc=0.0, scale=std, size=len(t))
    df['phi']+=noise

    return df

def check(X, estimator):
    estimator.fit(X=X)
    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

    assert_almost_equal(X['phi'].values, X_pred['phi'].values, decimal=3)
    assert estimator.score(X) > 0.999
    assert_almost_equal(estimator.parameters['zeta'], zeta, decimal=4)
    assert_almost_equal(estimator.parameters['omega0'], omega0, decimal=4)

def test_roll_decay_derivation_omega():

    estimator = RollDecay(fit_method='derivation', omega_regression=True)
    X = simulator(estimator=estimator)
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)

    check(X=X, estimator=estimator)

def test_roll_decay_integration_omega():

    estimator = RollDecay(fit_method='integration', omega_regression=True)
    X = simulator(estimator=estimator)
    check(X=X, estimator=estimator)

def test_roll_decay_integration_omega_negative():

    estimator = RollDecay(fit_method='integration', omega_regression=True)
    X = simulator_negative(estimator=estimator)
    check(X=X, estimator=estimator)

def test_roll_decay_derivation_no_omega():

    estimator = RollDecay(fit_method='derivation', omega_regression=False)
    X = simulator(estimator=estimator)
    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)

    check(X=X, estimator=estimator)

def test_roll_decay_integration_no_omega():

    estimator = RollDecay(fit_method='integration', omega_regression=False)
    X = simulator(estimator=estimator)
    check(X=X, estimator=estimator)

def test_roll_decay_integration_sparse():

    estimator = RollDecay(fit_method='integration', omega_regression=True)
    X = simulator_sparse(estimator=estimator)

    estimator.fit(X=X)
    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

def test_roll_decay_integration_noise():

    estimator = RollDecay(fit_method='integration', omega_regression=True)
    X = simulator_noise(estimator=estimator)

    estimator.fit(X=X)
    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

def test_load():

    estimator = RollDecay(fit_method='integration', omega_regression=False)
    X = simulator(estimator=estimator)

    data = {
        'omega0' : omega0,
        'zeta' : zeta,
    }

    estimator2 = RollDecay.load(data=data, X=X)
    X_pred = estimator2.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='loaded prediction')
    plt.show()

def test_load_score():

    estimator = RollDecay(fit_method='integration', omega_regression=False)
    X = simulator(estimator=estimator)

    data = {
        'omega0' : omega0,
        'zeta' : zeta,
    }

    estimator2 = RollDecay.load(data=data, X=X)
    assert estimator2.score(X=X) > 0.9999999999999





import numpy as np
import pandas as pd
from scipy.integrate import odeint

def roll_decay_time_step(states, t, d, omega0, zeta):
    # states:
    # [phi,phi1d]

    phi_old = states[0]
    p_old = states[1]

    phi1d = p_old
    phi2d = calculate_acceleration(d=d, omega0=omega0,phi1d=p_old, phi=phi_old, zeta=zeta)

    d_states_dt = np.array([phi1d, phi2d])

    return d_states_dt

def simulate(t :np.ndarray, phi0 :float, phi1d0 :float,omega0:float, d:float, zeta:float)->pd.DataFrame:
    """
    Simulate a roll decay test using the quadratic method.
    :param t: time vector to be simulated [s]
    :param phi0: initial roll angle [rad]
    :param phi1d0: initial roll speed [rad/s]
    :param omega0: roll natural frequency[rad/s]
    :param d: quadratic roll damping [-]
    :param zeta:linear roll damping [-]
    :return: pandas data frame with time series of 'phi' and 'phi1d'
    """

    states0 = [phi0, phi1d0]
    args = (
        d,
        omega0,
        zeta,
    )
    states = odeint(func=roll_decay_time_step, y0=states0, t=t, args=args)

    df = pd.DataFrame(index=t)

    df['phi'] = states[:, 0]
    df['phi1d'] = states[:,1]

    return df



from rolldecayestimators import equations
from rolldecayestimators.substitute_dynamic_symbols import lambdify
import sympy as sp
from rolldecayestimators.symbols import *

#calculate_acceleration_linear = lambdify(equations.acceleration_equation_linear.rhs)
#calculate_velocity_linear = lambdify(equations.velocity_equation_linear.rhs)




import sympy as sp
from rolldecayestimators.substitute_dynamic_symbols import lambdify
from rolldecayestimators import equations
from rolldecayestimators import symbols

B44_lambda = lambdify(sp.solve(equations.B44_hat_equation_quadratic,symbols.B_44_hat)[0])
B_1_hat_lambda = lambdify(sp.solve(equations.B_1_hat_equation, symbols.B_1_hat)[0])
B_e_hat_lambda = lambdify(sp.solve(equations.B_e_hat_equation, symbols.B_e_hat)[0])
B_2_hat_lambda = lambdify(sp.solve(equations.B_2_hat_equation, symbols.B_2_hat)[0])

omega0_lambda = lambdify(sp.solve(equations.omega0_hat_equation,symbols.omega_hat)[0])

omega = omega_from_hat = lambdify(sp.solve(equations.omega0_hat_equation,symbols.omega0)[0])
omega_hat = omega_to_hat = lambdify(sp.solve(equations.omega0_hat_equation,symbols.omega_hat)[0])


B_e_lambda = lambdify(sp.solve(equations.B_e_equation, symbols.B_e)[0])
B_e_lambda_cubic = lambdify(sp.solve(equations.B_e_equation_cubic, symbols.B_e)[0])

B44_hat_equation = equations.B44_hat_equation.subs(symbols.B_44,symbols.B)
B_hat_lambda=lambdify(sp.solve(B44_hat_equation,symbols.B_44_hat)[0])
B_to_hat_lambda=lambdify(sp.solve(B44_hat_equation,symbols.B_44_hat)[0])
B_from_hat_lambda=lambdify(sp.solve(B44_hat_equation,symbols.B)[0])

A44_lambda = lambdify(equations.A44)

omega0_linear_lambda = lambdify(sp.solve(equations.omega0_eq, symbols.omega0)[0])




"""
This is a module to be used as a reference for building other modules
"""
import numpy as np
import pandas as pd
from scipy.integrate import odeint
from sklearn.utils.validation import check_is_fitted
import inspect
from scipy.optimize import least_squares
import matplotlib.pyplot as plt
from rolldecayestimators import measure as measure
from sklearn.metrics import r2_score

from rolldecayestimators.substitute_dynamic_symbols import lambdify
from rolldecayestimators.symbols import *
from rolldecayestimators.estimator import RollDecay


class DirectEstimator(RollDecay):
    """ A template estimator to be used as a reference implementation.

    For more information regarding how to build your own estimator, read more
    in the :ref:`User Guide <user_guide>`.

    Parameters
    ----------
    bounds : dict, default=None
        Boundaries for the parameters expressed as dict:
        Ex: {
            'zeta':(-np.inf, np.inf),
            'd':(0,42),
            }

    fit_method : str, default='integration'
        The fitting method could be either 'integration' or 'derivation'
        'integration' means that the diff equation is solved with ODE integration.
        The curve_fit runs an integration for each set of parameters to get the best fit.

        'derivation' means that the diff equation os solved by first calculating the 1st and 2nd derivatives numerically.
        The derivatives are then inserted into the diff equation and the best fit is found.


    """

    # Defining the diff equation for this estimator:
    rhs = -phi_dot_dot / (omega0 ** 2) - 2 * zeta / omega0 * phi_dot - d * sp.Abs(phi_dot) * phi_dot / (omega0 ** 2)
    roll_diff_equation = sp.Eq(lhs=phi, rhs=rhs)
    acceleration = sp.Eq(lhs=phi, rhs=sp.solve(roll_diff_equation, phi.diff().diff())[0])
    functions = {'acceleration':lambdify(acceleration.rhs)}

    @classmethod
    def load(cls, omega0:float, d:float, zeta:float, X=None):
        """
        Load data and parameters from an existing fitted estimator

        Parameters
        ----------
        omega0 :
            roll frequency [rad/s]
        d : nondimensional linear damping
        zeta :
            nondimensional quadratic damping
        X : pd.DataFrame
            DataFrame containing the measurement that this estimator fits (optional).
        Returns
        -------
        estimator
            Loaded with parameters from data and maybe also a loaded measurement X
        """
        data={
            'd':d,
            'zeta':zeta,
            'omega0':omega0,
        }

        return super(cls, cls)._load(data=data, X=X)

    def calculate_amplitudes_and_damping(self):
        if hasattr(self,'X'):
            if not self.X is None:
                self.X_amplitudes=measure.calculate_amplitudes_and_damping(X=self.X)

                if self.is_fitted_:
                    X_pred = self.predict(X=self.X)
                    self.X_pred_amplitudes = measure.calculate_amplitudes_and_damping(X=X_pred)


    @staticmethod
    def calculate_damping(X_amplitudes):

        df_decrements = pd.DataFrame()

        for i in range(len(X_amplitudes) - 1):
            s1 = X_amplitudes.iloc[i]
            s2 = X_amplitudes.iloc[i + 1]

            decrement = s1 / s2
            decrement.name = s1.name
            df_decrements = df_decrements.append(decrement)

        df_decrements['zeta_n'] = 1 / (2 * np.pi) * np.log(df_decrements['phi'])

        df_decrements['zeta_n'] *= 2  # !!! # Todo: Where did this one come from?



        X_amplitudes_new = X_amplitudes.copy()
        X_amplitudes_new = X_amplitudes_new.iloc[0:-1].copy()
        X_amplitudes_new['zeta_n'] = df_decrements['zeta_n'].copy()
        X_amplitudes_new['B_n'] = 2*X_amplitudes_new['zeta_n']  # [Nm*s]

        return X_amplitudes_new


    def measure_error(self, X):
        y_true, y_pred = self.true_and_prediction(X=X)
        return y_pred - y_true

    def score(self, X=None, y=None, sample_weight=None):
        """
        Return the coefficient of determination R_b^2 of the prediction.

        The coefficient R_b^2 is defined as (1 - u/v), where u is the residual sum of squares
        ((y_true - y_pred) ** 2).sum() and v is the total sum of squares ((y_true - y_true.mean()) ** 2).sum().
        The best possible score is 1.0 and it can be negative (because the model can be arbitrarily worse).
        A constant model that always predicts the expected value of y, disregarding the input features,
        would get a R_b^2 score of 0.0.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Test samples. For some estimators this may be a precomputed kernel matrix or a list of generic
            objects instead, shape = (n_samples, n_samples_fitted), where n_samples_fitted is the number of samples
            used in the fitting for the estimator.

        y : Dummy not used

        sample_weight : Dummy

        Returns
        -------
        score : float
            R_b^2 of self.predict(X) wrt. y.

        """

        y_true, y_pred = self.true_and_prediction(X=X)

        #sample_weight = np.abs(y_true)
        #return r2_score(y_true=y_true, y_pred=y_pred,sample_weight=sample_weight)
        return r2_score(y_true=y_true, y_pred=y_pred)

    def calculate_average_linear_damping(self,phi_a=None):
        """
        Calculate the average linear damping
        In line with Himeno this equivalent linear damping is calculated as
        Parameters
        ----------
        phi_a : float, default = None
            linearize around this average angle [rad]
            phi_a is calculated based on data if None

        Returns
        -------
        average_linear_damping

        """
        check_is_fitted(self, 'is_fitted_')

        zeta = self.parameters['zeta']
        d = self.parameters.get('d',0)

        if phi_a is None:
            phi_a = self.X[self.phi_key].abs().mean()

        return zeta + 4/(3*np.pi)*d*phi_a

    def plot_fit(self, ax=None, include_model_test=True, label=None, **kwargs):

        check_is_fitted(self, 'is_fitted_')

        if ax is None:
            fig,ax = plt.subplots()

        df = self.predict(X=self.X)
        df['phi_deg'] = np.rad2deg(df['phi'])

        if include_model_test:
            X = self.X.copy()
            X['phi_deg'] = np.rad2deg(X['phi'])
            X.plot(y=r'phi_deg', ax=ax, label='Model test')

        if not label:
            label = 'fit'
        df.plot(y=r'phi_deg', ax=ax, label=label, style='--',**kwargs)

        ax.legend()
        ax.set_xlabel(r'Time [s]')
        ax.set_ylabel(r'$\Phi$ [deg]')

    def plot_error(self,X=None, ax=None, **kwargs):
        check_is_fitted(self, 'is_fitted_')

        if ax is None:
            fig, ax = plt.subplots()

        error = self.measure_error(X=X)

        ax.plot(self.X.index, error, label=self.__repr__(), **kwargs)
        ax.legend()
        ax.set_xlabel('Time [s]')
        ax.set_ylabel('error: phi_pred - phi_true [rad]')

    def plot_peaks(self, ax=None, **kwargs):
        check_is_fitted(self, 'is_fitted_')

        if ax is None:
            fig,ax = plt.subplots()

        #self.X.plot(y='phi', ax=ax)

        self.X_amplitudes.plot(y='phi', ax=ax, **kwargs)

        #ax.plot([np.min(self.X.index),np.max(self.X.index)],[0,0],'m-')
        ax.set_title('Peaks')

    def plot_velocity(self, ax=None):
        check_is_fitted(self, 'is_fitted_')

        if ax is None:
            fig,ax = plt.subplots()

        self.X.plot(y='phi1d', ax=ax)
        self.X_amplitudes.plot(y='phi1d', ax=ax, style='r.')
        ax.plot([np.min(self.X.index), np.max(self.X.index)], [0, 0], 'm-')
        ax.set_title('Velocities')

    def plot_amplitude(self, ax=None, include_model_test=True):

        if ax is None:
            fig,ax = plt.subplots()

        if not hasattr(self,'X_amplitudes'):
            self.calculate_amplitudes_and_damping()

        if include_model_test:
            X_amplitudes=self.X_amplitudes.copy()
            X_amplitudes['phi_a']=np.rad2deg(X_amplitudes['phi_a'])
            X_amplitudes.plot(y='phi_a', style='o', label='Model test', ax=ax)

        if hasattr(self,'X_pred_amplitudes'):
            label = self.__repr__()
            X_pred_amplitudes = self.X_pred_amplitudes.copy()
            X_pred_amplitudes['phi_a'] = np.rad2deg(X_pred_amplitudes['phi_a'])
            X_pred_amplitudes.plot(y='phi_a', label=label, ax=ax)

    def plot_damping(self, ax=None, include_model_test=True,label=None, **kwargs):

        if ax is None:
            fig, ax = plt.subplots()

        plot = None
        if include_model_test:
            X_amplitudes = measure.calculate_amplitudes_and_damping(X=self.X)
            X_amplitudes['phi_a'] = np.rad2deg(X_amplitudes['phi_a'])
            plot = X_amplitudes.plot(x='phi_a', y='B_n', style='o', label='Model test', ax=ax, **kwargs)

        if self.is_fitted_:
            if not label:
                label = self.__repr__()

            X_pred = self.predict(X=self.X)
            X_pred_amplitudes = measure.calculate_amplitudes_and_damping(X=X_pred)
            X_pred_amplitudes['phi_a'] = np.rad2deg(X_pred_amplitudes['phi_a'])

            if plot is None:
                color=None
            else:
                line = plot.axes.get_lines()[-1]
                color = line.get_color()

            x = X_pred_amplitudes['phi_a']
            y = X_pred_amplitudes['B_n']
            ax.plot(x, y, color=color, label=label, **kwargs, lw=2)

        ax.set_xlabel(r'$\phi_a$ [deg]')
        ax.set_ylabel(r'$B$ [Nms]')
        ax.legend()

    def plot_omega0(self,ax=None, include_model_test=True, label=None, **kwargs):

        if ax is None:
            fig,ax = plt.subplots()

        if not hasattr(self, 'X_amplitudes'):
            self.calculate_amplitudes_and_damping()

        plot = None
        if include_model_test:

            X_amplitudes = self.X_amplitudes.copy()
            X_amplitudes['omega0_norm'] = X_amplitudes['omega0']/self.omega0
            X_amplitudes['phi_a_deg'] = np.rad2deg(X_amplitudes['phi_a'])
            plot = X_amplitudes.plot(x='phi_a_deg', y='omega0_norm', style='o', label='Model test', ax=ax, **kwargs)

        if hasattr(self, 'X_pred_amplitudes'):
            if not label:
                label = self.__repr__()

            if plot is None:
                color = None
            else:
                line = plot.axes.get_lines()[-1]
                color = line.get_color()

            x = np.rad2deg(self.X_pred_amplitudes['phi_a'])
            y = self.X_pred_amplitudes['omega0']/self.omega0

            ax.plot(x, y, color=color, label=label, **kwargs, lw=2)

        ax.set_xlabel(r'$\phi_a$ [deg]')
        ax.set_ylabel(r'$\frac{\omega_0^N}{\omega_0}$ [-]')
        ax.legend()

    def plot_fft(self, ax=None):
        check_is_fitted(self, 'is_fitted_')

        if ax is None:
            fig,ax = plt.subplots()

        frequencies, dft = self.fft(self.X['phi'])
        omega=2*np.pi*frequencies

        omega0 = self.fft_omega0(frequencies=frequencies, dft=dft)
        index = np.argmax(np.abs(dft))
        ax.plot(omega, dft)
        ax.plot(omega0,dft[index],'ro')



import pytest

import pandas as pd
import numpy as np
from numpy.testing import assert_almost_equal
from rolldecayestimators.direct_estimator import DirectEstimator
from rolldecayestimators.transformers import CutTransformer
import matplotlib.pyplot as plt

@pytest.fixture
def d():
    yield 0.076

@pytest.fixture
def omega0():
    T0 = 20.0
    yield 2.0*np.pi/T0

@pytest.fixture
def zeta():
    yield 0.044

@pytest.fixture
def df_roll_decay(omega0, d, zeta):

    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 120, 0.01)
    estimator = DirectEstimator()
    estimator.parameters = {
        'omega0':omega0,
        'd':d,
        'zeta':zeta,
    }

    yield estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)


def test_cutter(df_roll_decay):

    cut_transformer = CutTransformer(phi_max=np.deg2rad(5))
    cut_transformer.fit(df_roll_decay)
    X = cut_transformer.transform(df_roll_decay)

    fig,ax=plt.subplots()
    df_roll_decay.plot(y='phi', ax=ax, label='raw')
    X.plot(y='phi', ax=ax, label='cut', style='--')
    plt.show()





"""
The classes in this module extends the Sympy Symbol class, to also carry description, unit and bis denominator
"""

import sympy as sp
from collections import OrderedDict

# my custom class with description attribute
class Symbol(sp.Symbol):
    def __new__(self, name, description='', unit=''):
        obj = sp.Symbol.__new__(self, name)
        obj.description = description
        obj.unit = unit
        return obj


class Coefficient(Symbol):
    def __new__(self, name, description='coefficient', unit=''):
        obj = super().__new__(self, name, description=description, unit=unit)
        return obj

class BisSymbol(Symbol):
    def __new__(self, name,parent_SI_symbol, description='', unit=''):
        obj = super().__new__(self, name, description=description, unit=unit)
        obj.parent_SI_symbol = parent_SI_symbol
        return obj

class Bis(Symbol):

    def __new__(self, name, denominator, description='', unit=''):
        obj = super().__new__(self, name, description=description, unit=unit)
        obj.denominator = denominator

        bis_name = "%s''%s" % (name[0], name[1:])
        obj.bis = BisSymbol(bis_name,parent_SI_symbol = obj, description=description, unit=unit)
        obj.bis_eq = sp.Eq(lhs=obj.bis, rhs=obj / denominator)

        return obj

def expand_bis(equation:sp.Eq):
    """
    Remove all bis symbols from an expression by substituting with the corresponding bis equation
    :param equation: sympy equation
    :return: new equation WITHOUT bis symbols.
    """

    assert isinstance(equation,sp.Eq)
    symbols = equation.lhs.free_symbols | equation.rhs.free_symbols
    subs = []
    for symbol in symbols:
        if isinstance(symbol,BisSymbol):
            subs.append((symbol,symbol.parent_SI_symbol.bis_eq.rhs))

    expanded_equation = equation.subs(subs)
    return expanded_equation

def reduce_bis(equation:sp.Eq):
    """
    Replace symbols with corresponding bis symbols from an expression by substituting with the corresponding bis equation
    :param equation: sympy equation
    :return: new equation WITH bis symbols.
    """

    assert isinstance(equation,sp.Eq)
    symbols = equation.lhs.free_symbols | equation.rhs.free_symbols
    subs = []
    for symbol in symbols:
        if isinstance(symbol,Bis):
            subs.append((symbol,sp.solve(symbol.bis_eq,symbol)[0]))

    reduced = equation.subs(subs)
    return reduced


def create_html_table(symbols:list):


    html = """
        <tr>
            <th>Variable</th>
            <th>Description</th> 
            <th>SI Unit</th> 
        </tr>
        """

    names = [symbol.name for symbol in symbols if isinstance(symbol, sp.Basic)]
    symbols_dict = {symbol.name:symbol for symbol in symbols if isinstance(symbol, sp.Basic)}

    for name in sorted(names):
        symbol = symbols_dict[name]

        if isinstance(symbol, Symbol):
            html_row = """
            <tr>
                <td>$%s$</td>
                <td>%s</td> 
                <td>%s</td> 
            </tr>
            """ % (sp.latex(symbol), symbol.description, symbol.unit)

            html += html_row

    html_table = """
    <table>
    %s
    </table>
    """ % html
    return html_table


import pytest

import pandas as pd
import numpy as np
from numpy.testing import assert_almost_equal
from rolldecayestimators.direct_estimator_cubic import EstimatorCubic, EstimatorQuadraticB, EstimatorLinear
import matplotlib.pyplot as plt
from rolldecayestimators.estimator import FitError

def simulate(t, phi0, phi1d0, **kwargs):

    estimator = EstimatorCubic.load(**kwargs)

    return estimator.simulate(t=t,phi0=phi0, phi1d0=phi1d0)

def check(X, estimator, parameters, decimal=2):

    if not estimator.is_fitted_:
        estimator.fit(X=X)

    assert estimator.result['success']
    X_pred = estimator.predict(X=X)
    fig, ax = plt.subplots()
    X.plot(y='phi', ax=ax, label='actual')
    X_pred.plot(y='phi', ax=ax, label='prediction')
    plt.show()

    assert_almost_equal(X['phi'].values, X_pred['phi'].values, decimal=decimal)
    assert estimator.score(X) > 0.999


    true_parameters = pd.Series(parameters)
    predicted = pd.Series(estimator.parameters, index=true_parameters.index)
    predicted.fillna(0, inplace=True)

    # Normalize with A_44:
    #true_parameters/=true_parameters['A_44']
    #predicted/=predicted['A_44']

    pd.testing.assert_series_equal(predicted.round(decimals=2), true_parameters)

    #for key,value in parameters.items():
    #    try:
    #        assert_almost_equal(estimator.parameters[key], value, decimal=decimal)
    #    except:
    #        raise ValueError('%s predicted:%f, true:%f' % (key, estimator.parameters[key], value))


def test_simulation():

    parameters={
        'B_1A':0.3,
        'B_2A':0.0,
        'B_3A':0.0,
        'C_1A':0.3,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.1)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    direct_estimator = EstimatorCubic.load(**parameters)

    direct_estimator.is_fitted_=True
    X_pred = direct_estimator.predict(X=X)
    assert_almost_equal(X['phi'].values, X_pred['phi'].values)

def test_fit_simualtion_cheat_p0():

    parameters={
        'B_1A':0.7,
        'B_2A':0.0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    p0 = parameters

    direct_estimator = EstimatorCubic(fit_method='integration', p0=p0)
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion():

    parameters={
        'B_1A':0.7,
        'B_2A':0.0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    bounds = {
        'B_1A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_2A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_3A': (-np.inf, np.inf),  # Assuming only positive coefficients

    }

    direct_estimator = EstimatorCubic(fit_method='integration', bounds=bounds)
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_quadratic_damping():

    parameters={
        'B_1A':0.7,
        'B_2A':1.5,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    bounds = {
        'B_1A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_2A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_3A': (-np.inf, np.inf),  # Assuming only positive coefficients

    }

    direct_estimator = EstimatorCubic(fit_method='integration', bounds=bounds)
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_cubic_damping():

    parameters={
        'B_1A':0.7,
        'B_2A':1.5,
        'B_3A':5.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorCubic(fit_method='integration',maxfev=100, )
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_cubic_damping_not_converged():

    parameters={
        'B_1A':0.7,
        'B_2A':1.5,
        'B_3A':5.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorCubic(fit_method='integration',maxfev=1, )
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)

    with pytest.raises(FitError):
        check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_quadratic_stiffness():

    parameters={
        'B_1A':0.7,
        'B_2A':0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':20.0,
        'C_5A':0.0,
    }

    bounds = {
        'B_1A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_2A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_3A': (-np.inf, np.inf),  # Assuming only positive coefficients

    }

    direct_estimator = EstimatorCubic(fit_method='integration', bounds=bounds)
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_cubic_stiffness():

    parameters={
        'B_1A':0.7,
        'B_2A':0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':20.0,
        'C_5A':1000.0,
    }

    bounds = {
        'B_1A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_2A': (-np.inf, np.inf),  # Assuming only positive coefficients
        'B_3A': (-np.inf, np.inf),  # Assuming only positive coefficients

    }

    direct_estimator = EstimatorCubic(fit_method='integration',bounds=bounds)
    phi0 = np.deg2rad(10)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_fit_simualtion_full_cubic():

    parameters={
        'B_1A':0.7,
        'B_2A':1.0,
        'B_3A':3.0,
        'C_1A':10.0,
        'C_3A':10.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorCubic(fit_method='integration')
    phi0 = np.deg2rad(20)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)


def test_simulation_quadratic():

    parameters = {
        'B_1A': 100,
        'B_2A': 20,
        'C_1A': 100,

    }

    direct_estimator = EstimatorQuadraticB.load(**parameters)


    phi0 = np.deg2rad(20)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)

    direct_estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

def test_fit_simulation_quadratic():

    parameters={
        'B_1A':0.7,
        'B_2A':1.0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorQuadraticB(fit_method='integration')
    phi0 = np.deg2rad(20)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_simulation_linear():

    parameters = {
        'B_1A': 100,
        'C_1A': 100,

    }

    direct_estimator = EstimatorLinear.load(**parameters)


    phi0 = np.deg2rad(20)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)

    direct_estimator.simulate(t=t, phi0=phi0, phi1d0=phi1d0)

def test_fit_simulation_linear():

    parameters={
        'B_1A':0.7,
        'B_2A':0.0,
        'B_3A':0.0,
        'C_1A':10.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorLinear(fit_method='integration')

    phi0 = np.deg2rad(20)
    phi1d0 = 0
    t = np.arange(0, 10, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    check(X=X, estimator=direct_estimator, parameters=parameters)

def test_result_for_database():

    parameters={
        'B_1A':0.3,
        'B_2A':0.0,
        'B_3A':0.0,
        'C_1A':4.0,
        'C_3A':0.0,
        'C_5A':0.0,
    }

    direct_estimator = EstimatorLinear(fit_method='integration')
    phi0 = np.deg2rad(5)
    phi1d0 = 0
    t = np.arange(0, 20, 0.01)
    X = simulate(t=t, phi0=phi0, phi1d0=phi1d0, **parameters)

    X['phi2d'] = np.gradient(X['phi1d'].values, X.index.values)
    direct_estimator.fit(X=X)

    rho=1000
    g=9.81
    meta_data = {
        'GM':1,
        'Volume':1/rho,
        'rho':rho,
        'g':g,
    }

    check(X=X, estimator=direct_estimator, parameters=parameters)
    s = direct_estimator.result_for_database(meta_data=meta_data)

    assert_almost_equal(s['omega0'],np.sqrt(parameters['C_1A']))

    g=9.81
    m=meta_data['Volume']*rho
    C = meta_data['GM']*g*m
    A_44 = C/s['omega0']**2
    assert_almost_equal(s['A_44'],A_44)
    B_1 = s['B_1A']*A_44
    assert_almost_equal(s['B_1'],B_1)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from rolldecayestimators.simplified_ikeda import calculate_roll_damping, SimplifiedIkedaInputError

def variate_ship(ship, key, changes):
    N = len(changes)
    data = np.tile(ship.values, (N, 1))

    df = pd.DataFrame(data=data, columns=ship.index)
    variations = changes * ship[key]
    df[key] = variations
    df.index = df[key].copy()

    return df

def calculate_variation(df, catch_error=False, limit_inputs=False, verify_input=True):
    result = df.apply(func=calculate, catch_error=catch_error, limit_inputs=limit_inputs, verify_input=verify_input, axis=1)
    return result

def plot_variation(ship, key='lpp', changes=None, ax=None, catch_error=False, plot_change_factor=True):

    if changes is None:
        N = 30
        changes = np.linspace(0.5, 1.5, N)

    df = variate_ship(ship=ship, key=key, changes=changes)
    result = calculate_variation(df=df, catch_error=catch_error)
    result[key] = df[key].copy()

    ax = _plot_result(ship=ship, result=result, key=key, changes=changes, ax=ax, plot_change_factor=plot_change_factor)
    return ax

def _plot_result(ship, result, key, changes, plot_change_factor=True, ax=None):

    if ax is None:
        fig, ax = plt.subplots()

    if plot_change_factor:
        result['change factor'] = changes
        result.plot(x='change factor', ax=ax)
    else:
        result.plot(x=key, ax=ax)

    ax.set_title('Variation of %s: %0.3f' % (key, ship[key]))
    return ax

def calculate(row, catch_error=False, limit_inputs=False, verify_input=True, **kwargs):
    LPP = row.lpp
    Beam = row.beam
    DRAFT = row.DRAFT

    PHI = row.phi_max
    lBK = row.BKL
    bBK = row.BKB
    OMEGA = row.omega0
    OG = (-row.kg + DRAFT)
    CB = row.CB
    CMID = row.A0
    V = row.V

    s = pd.Series()
    try:
        B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT = calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, lBK, bBK,
                                                                    OMEGA, DRAFT, V=V ,limit_inputs=limit_inputs,
                                                                            verify_input=verify_input, **kwargs)
    except SimplifiedIkedaInputError:
        if catch_error:
            return s
        else:
            raise

    s['B_44_hat'] = B44HAT
    s['B_F_hat'] = BFHAT
    s['B_W_hat'] = BWHAT
    s['B_E_hat'] = BEHAT
    s['B_BK_hat'] = BBKHAT
    s['B_L_hat'] = BLHAT
    return s


import pytest
from rolldecayestimators.simplified_ikeda import calculate_roll_damping, _calculate_roll_damping, limits_kawahara, verify_inputs, SimplifiedIkedaInputError
import rolldecayestimators.ikeda_simple  ## Peter Piehl implementation
import numpy as np
from numpy.testing import assert_almost_equal

"""
L / B = 6.0
B / d = 4.0,
Cb = 0.65,
Cm = 0.98,
φa = 10,
BKB / B = 0.025
BKL / Lpp = 0.2
"""
L_div_B = 6.0
BD = 4.0
CB = 0.65
CMID = 0.98
PHI = 10
BBKB = 0.025
LBKL = 0.2
OGD = -0.2
OMEGAHAT = 0.6
LPP = 300

def test_calculate_roll_damping():


    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * LBKL
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))

    OG = DRAFT * OGD

    B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT = calculate_roll_damping(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,DRAFT)
    assert_almost_equal(B44HAT,0.010156148773301035)


def test_calculate_roll_damping_subfunction():

    Beam = BRTH= LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * LBKL
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))

    OG = DRAFT * OGD
    TW = 2*np.pi/OMEGA

    B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT = calculate_roll_damping(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,DRAFT)
    B44HAT2, BFHAT2, BWHAT2, BEHAT2, BBKHAT2 = _calculate_roll_damping(LPP, BRTH, CB, CMID, OGD, PHI, LBKL, BBKB, OMEGA,
                           DRAFT, BD, OMEGAHAT)

    assert B44HAT==B44HAT2
    assert BFHAT == BFHAT2
    assert BWHAT == BWHAT2
    assert BEHAT == BEHAT2
    assert BBKHAT == BBKHAT2

def test_verify_input():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * LBKL
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT * OGD

    verify_inputs(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,
                           DRAFT)


def test_verify_input_fail1():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * LBKL
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT * OGD

    CB_fail=0.45
    with pytest.raises(SimplifiedIkedaInputError):

        verify_inputs(LPP,Beam,CB_fail,CMID,OG,PHI,lBK,bBK,OMEGA,
                           DRAFT)

def test_verify_input_fail2():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * 0.44
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT * OGD

    with pytest.raises(SimplifiedIkedaInputError):

        verify_inputs(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,
                           DRAFT)

def test_limit_input():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * 0.44
    bBK = Beam * BBKB
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT * OGD

    with pytest.raises(SimplifiedIkedaInputError):

        calculate_roll_damping(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,DRAFT)

    calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, lBK, bBK, OMEGA, DRAFT, limit_inputs=True)

def test_limit_input2():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * 0.30
    bBK = Beam * 0.001
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT * OGD

    with pytest.raises(SimplifiedIkedaInputError):

        calculate_roll_damping(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,DRAFT)

    calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, lBK, bBK, OMEGA, DRAFT, limit_inputs=True)

def test_limit_input3():
    Beam = LPP/L_div_B
    DRAFT = Beam / BD

    lBK = LPP * 0.30
    bBK = Beam * 0.002
    OMEGA = OMEGAHAT/(np.sqrt(Beam / 2 / 9.81))
    OG = DRAFT*0.25

    with pytest.raises(SimplifiedIkedaInputError):

        calculate_roll_damping(LPP,Beam,CB,CMID,OG,PHI,lBK,bBK,OMEGA,DRAFT)

    calculate_roll_damping(LPP, Beam, CB, CMID, OG, PHI, lBK, bBK, OMEGA, DRAFT, limit_inputs=True)


def random_value(limits):
    return np.random.random()*(limits[1]-limits[0])+limits[0]

def random_values():
    LPP=1.0
    Beam=1.0
    DRAFT=Beam/2.5
    value={
        'LPP':LPP,
        'Beam':Beam,
        'DRAFT':DRAFT,
        'CB':random_value(limits_kawahara['CB']),
        'CMID':random_value(limits_kawahara['CMID']),
        'OG':DRAFT*random_value(limits_kawahara[r'OG/d']),
        'PHI':PHI,
        'BKL':LPP*random_value(limits_kawahara[r'lBk/LPP']),
        'BKB':Beam*random_value(limits_kawahara[r'bBk/B']),
        'OMEGA':random_value(limits_kawahara['OMEGA_hat'])/(np.sqrt(Beam / 2 / 9.81)),

    }
    return value



def test_peter_piehl_implementation():
    np.random.seed(seed=1)

    ikeda = rolldecayestimators.ikeda_simple.Ikeda()  # Peter Piehl

    for i in range(1000):
        r = random_values()
        B44HAT, BFHAT, BWHAT, BEHAT, BBKHAT, BLHAT = calculate_roll_damping(**r)

        # Peter Piehl:

        r2 = {}
        r2['CMID']=r['CMID']
        r2['TW']=(2*np.pi)/r['OMEGA']
        r2['LPP']=r['LPP']
        r2['BRTH']=r['Beam']
        r2['DRAFT']=r['DRAFT']
        r2['CB']=r['CB']
        r2['OG']=r['OG']
        r2['PHI']=r['PHI']
        r2['BKCOMP']=(r['BKL']>0)
        r2['BKL']=r['BKL']
        r2['BKB']=r['BKB']

        ikeda.setPara(r2)
        ikeda.ikedaMethod()

        assert_almost_equal(B44HAT,ikeda.B44HAT)
        assert_almost_equal(BFHAT, ikeda.BFHAT)
        assert_almost_equal(BWHAT, ikeda.BWHAT)
        assert_almost_equal(BEHAT, ikeda.BEHAT)
        assert_almost_equal(BBKHAT, ikeda.BBKHAT)





import numpy as np
import pytest
from rolldecayestimators import tests
import rolldecayestimators
import pandas as pd
import os
from rolldecayestimators import ikeda_speed as ikeda
from numpy import pi, sqrt
from numpy.testing import assert_almost_equal
from matplotlib import pyplot as plt

@pytest.fixture
def lewis_coefficients():
    file_path = os.path.join(rolldecayestimators.path,'faust_lewis.csv')
    df = pd.read_csv(file_path, sep=';')

    yield df

def test_eddy():

    N = 21
    lpp=100.0
    OG=5.5
    R=2
    d=1
    wE=0.2
    fi_a=np.deg2rad(3)
    xs = np.linspace(0,lpp, N)
    B = T = S = np.ones(N)
    a, a_1, a_3, sigma_s, H = ikeda.calculate_sectional_lewis(B_s=B, T_s=T, S_s=S)

    B_E = ikeda.eddy(bwl=B, a_1=a_1, a_3=a_3, sigma=sigma_s, xs=xs, H0=H, Ts=T, OG=OG, R=R, wE=wE, fi_a=fi_a)

def test_eddy_faust(lewis_coefficients):
    """
    Reproduction of Carl-Johans Matlab implementation for Faust.
    Parameters
    ----------
    lewis_coefficients

    Returns
    -------

    """

    lc=lewis_coefficients

    T = 27.6
    wE = 2*pi*1/T  #  circular frequency
    d    =   9.5  # Draught of hull [m]
    vcg  =   14.9  # roll axis (vertical centre of gravity) [m]
    OG = -1 * (vcg - d)   # distance from roll axis to still water level
    fi_a =   10*pi/180  # roll amplitude !!rad??
    R = 5  # Bilge Radis
    B_E = ikeda.eddy(bwl=lc['bwl'], a_1=lc['a1'], a_3=lc['a3'], sigma=lc['sigma'], xs=lc['x'], H0=lc['H'], Ts=lc['Ts'],
                     OG=OG, R=R, wE=wE, fi_a=fi_a)

    #assert_almost_equal(actual=B_E, desired=1175062.2691943)

    ScaleF =  1#/29.565                  # Scale Factor [-]
    Cb   =   0.61                        # Block coeff
    L    =   220*ScaleF                  # Length
    vcg  =   14.9*ScaleF                 # roll axis (vertical centre of gravity) [m]
    B    =   32.26*ScaleF                # Breadth of hull [m]
    d    =   9.5*ScaleF                  # Draught of hull [m]
    g    =   9.81

    ra   = 1025                   # density of water

    disp = L*B*d*Cb  # Displacement
    ND_factorB = sqrt(B / (2 * g)) / (ra * disp * (B**2))

    w_hat = np.linspace(0,1,100)
    w = sqrt(2) * w_hat / sqrt(B / g)
    B_E = ikeda.eddy(bwl=lc['bwl'], a_1=lc['a1'], a_3=lc['a3'], sigma=lc['sigma'], xs=lc['x'], H0=lc['H'], Ts=lc['Ts'],
                     OG=OG, R=R, wE=w, fi_a=fi_a)

    B_E_hat = B_E*ND_factorB
    fig,ax=plt.subplots()
    ax.plot(w_hat, B_E_hat)
    plt.show()



def test_calculate_sectional_lewis():
    N=21
    B=T=S=np.ones(N)
    a, a_1, a_3, sigma_s, H = ikeda.calculate_sectional_lewis(B_s=B, T_s=T, S_s=S)